# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

from gitfourchette.qt import *
from gitfourchette.toolbox import *


class CodeGutter(QWidget):
    """
    Generic gutter for CodeView
    """
    # Inspired by https://doc.qt.io/qt-6.2/qtwidgets-widgets-codeeditor-example.html

    lineClicked = Signal(QPoint)
    lineShiftClicked = Signal(QPoint)
    lineDoubleClicked = Signal(QPoint)
    selectionMiddleClicked = Signal()

    codeView: QPlainTextEdit

    def __init__(self, parent: QPlainTextEdit):
        super().__init__(parent)
        self.codeView = parent

        cursorDpr = 1 if (FREEDESKTOP and not WAYLAND) else 4  # Qt doesn't seem to support cursor DPR on X11
        cursorPix = QPixmap(f"assets:icons/right_ptr@{cursorDpr}x")
        cursorPix.setDevicePixelRatio(cursorDpr)
        flippedCursor = QCursor(cursorPix, hotX=19, hotY=5)
        self.setCursor(flippedCursor)

        # Enable customContextMenuRequested signal
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

    def syncFont(self, codeFont: QFont):
        self.setFont(codeFont)

    def refreshMetrics(self):
        if type(self) is not CodeGutter:  # pragma: no cover
            raise NotImplementedError("CodeGutter subclasses must override this function")

    def calcWidth(self) -> int:
        if type(self) is not CodeGutter:  # pragma: no cover
            raise NotImplementedError("CodeGutter subclasses must override this function")

        return 0

    def onParentUpdateRequest(self, rect: QRect, dy: int):
        if dy != 0:
            self.scroll(0, dy)
        else:
            self.update(0, rect.y(), self.width(), rect.height())

    def sizeHint(self) -> QSize:
        return QSize(self.calcWidth(), 0)

    def wheelEvent(self, event: QWheelEvent):
        # Forward mouse wheel to parent widget
        self.parentWidget().wheelEvent(event)

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        # Double click to select clump of lines
        if event.button() == Qt.MouseButton.LeftButton:
            pos = event.position().toPoint()
            self.lineDoubleClicked.emit(pos)

    def mousePressEvent(self, event: QMouseEvent):
        if event.button() == Qt.MouseButton.LeftButton:
            pos = event.position().toPoint()
            if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                self.lineShiftClicked.emit(pos)
            else:
                self.lineClicked.emit(pos)

    def mouseReleaseEvent(self, event: QMouseEvent):
        if event.button() == Qt.MouseButton.MiddleButton:
            self.selectionMiddleClicked.emit()

    def mouseMoveEvent(self, event: QMouseEvent):
        if event.buttons() == Qt.MouseButton.LeftButton:
            pos = event.position().toPoint()
            self.lineShiftClicked.emit(pos)

    def paintBlocks(self, event: QPaintEvent, painter: QPainter, lineColor: QColor):
        # Set up colors
        palette = self.palette()
        themeBG = palette.color(QPalette.ColorRole.Base)  # standard theme background color
        if isDarkTheme(palette):
            gutterColor = themeBG.darker(105)
        else:
            gutterColor = themeBG.lighter(140)

        # Gather some metrics
        paintRect = event.rect()
        gutterRect = self.rect()
        rightEdge = gutterRect.width() - 1

        # Clip painting to QScrollArea viewport rect (don't draw beneath horizontal scroll bar)
        vpRect = self.codeView.viewport().rect()
        vpRect.setWidth(paintRect.width())  # vpRect is adjusted by gutter width, so undo this
        paintRect = paintRect.intersected(vpRect)
        painter.setClipRect(paintRect)

        # Draw background
        painter.fillRect(paintRect, gutterColor)

        # Draw vertical separator line
        painter.fillRect(rightEdge, paintRect.y(), 1, paintRect.height(), lineColor)

        block: QTextBlock = self.codeView.firstVisibleBlock()
        top = round(self.codeView.blockBoundingGeometry(block).translated(self.codeView.contentOffset()).top())
        bottom = top + round(self.codeView.blockBoundingRect(block).height())

        while block.isValid() and top <= paintRect.bottom():
            if block.isVisible() and bottom >= paintRect.top():
                yield block, top, bottom

            block = block.next()
            top = bottom
            bottom = top + round(self.codeView.blockBoundingRect(block).height())



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import re
from collections.abc import Iterable, Callable, Container
from html import escape

from gitfourchette.localization import *
from gitfourchette.qt import *

_elideMetrics: QFontMetrics | None = None

_naturalSortSplit = re.compile(r"(\d+)")

_stripAccelerators = re.compile(r"&(?!&)")

_titleLowercaseWords = {"a", "an", "and", "as", "but", "by", "in", "of", "on", "or", "the", "to"}


def toLengthVariants(pipeSeparatedString: str) -> str:
    return pipeSeparatedString.replace("|", "\x9C")


def getElideMetrics() -> QFontMetrics:
    # Cannot initialize _elideMetrics too early for Windows offscreen unit tests
    global _elideMetrics
    if _elideMetrics is None:
        _elideMetrics = QFontMetrics(QFontDatabase.systemFont(QFontDatabase.SystemFont.GeneralFont))
    return _elideMetrics


def messageSummary(body: str, elision=" [\u2026]"):
    messageContinued = False
    message: str = body.strip()
    newline = message.find('\n')
    if newline > -1:
        messageContinued = newline < len(message) - 1
        message = message[:newline]
        if messageContinued:
            message += elision
    return message, messageContinued


def escamp(text: str) -> str:
    """ Sanitize ampersands in user strings for QLabel. """
    return text.replace('&', '&&')


def paragraphs(*strings: str) -> str:
    """ Return a string of HTML "P" tags surrounding each argument. """
    inner = "</p><p>".join(strings)
    return f"<p>{inner}</p>"


def tquo(text: str) -> str:
    """ Quote plain text with language-dependent typographic quotes. """
    return _("“{0}”").format(text)


def hquo(text: str) -> str:
    """ Quote text, HTML-safe. """
    return tquo(escape(text))


def hquoe(text: str) -> str:
    """ Quote and elide text, HTML-safe. """
    return tquo(escape(elide(text)))


def bquo(text: str) -> str:
    """ Quote text, bold HTML-safe. """
    return tquo(btag(text))


def bquoe(text: str) -> str:
    """ Quote and elide text, bold HTML-safe. """
    return tquo(btag(elide(text)))


def lquo(text: str) -> str:
    """ Quote text, ampersand-safe (for Qt labels). """
    return tquo(escamp(text))


def lquoe(text: str) -> str:
    """ Quote and elide text, ampersand-safe (for Qt labels). """
    return tquo(escamp(elide(text)))


def tquoe(text: str) -> str:
    """ Quote and elide plain text. """
    return tquo(elide(text))


def btag(text: str) -> str:
    """ Surround text with HTML "B" tags, HTML-safe. """
    return f"<b>{escape(text)}</b>"


def stripHtml(markup: str):
    return QTextDocumentFragment.fromHtml(markup).toPlainText()


def stripAccelerators(text: str):
    return _stripAccelerators.sub("", text)


def elide(text: str, mode: Qt.TextElideMode = Qt.TextElideMode.ElideMiddle, ems: int = 20):
    metrics = getElideMetrics()
    maxWidth = metrics.horizontalAdvance(ems * 'M')
    return metrics.elidedText(text, mode, maxWidth)


def clipboardStatusMessage(text: str):
    n = 1 + text.count('\n')
    if n == 1:
        return _("{0} copied to clipboard.", tquoe(text))
    else:
        return _("{n} lines copied to clipboard.", n=n)


def ulify(items: Iterable[str], limit: int = 10, prefix="", suffix="", moreText=""):
    n = 0
    text = "<ul>"

    for item in items:
        if limit < 0 or n < limit:
            text += f"\n<li>{prefix}{item}{suffix}</li>"
        n += 1

    if n == 0:
        return ""

    if 0 <= limit < n:
        unlisted = n - limit
        if not moreText:
            moreText = _("…and {0} more")
        moreText = moreText.format(unlisted)
        text += f"\n<li>{prefix}<i>{moreText}</i>{suffix}</li>"

    text += "\n</ul>"
    return text


def toTightUL(items: Iterable[str], limit=10, moreText=""):
    return ulify(items, limit=limit, moreText=moreText)


def toRoomyUL(items: Iterable[str]):
    return ulify(items, -1, "<p>", "</p>")


def linkify(text, *mixedTargets: str | QUrl):
    targets = [target.toString() if isinstance(target, QUrl) else target
               for target in mixedTargets]

    assert all('"' not in target for target in targets)

    if "[" not in text:
        assert len(targets) == 1
        return f"<a href=\"{targets[0]}\">{text}</a>"

    for target in targets:
        assert "[" in text
        text = text.replace("[", f"<a href=\"{target}\">", 1).replace("]", "</a>", 1)

    return text


def tagify(text, *tags: str):
    def closingTag(tag: str):
        rtags = tag.split("<")[1:]
        rtags.append("")
        return "</".join(reversed(rtags))

    if "[" not in text:
        assert "]" not in text
        text = f"[{text}]"

    for tag in tags:
        text = text.replace("[", tag, 1).replace("]", closingTag(tag), 1)

    return text


def withUniqueSuffix(
        stem: str, reserved: Container[str] | Callable[[str], bool],
        start=2, stop=-1,
        ext="", suffixFormat="-{}"):
    # Test format first to catch any errors even if we don't enter the loop
    assert suffixFormat.format(1) != suffixFormat.format(2), "illegal suffixFormat"

    name = stem + ext
    i = start

    isTaken: Callable[[str], bool]
    if not callable(reserved):
        isTaken = reserved.__contains__
    else:
        isTaken = reserved

    while isTaken(name):
        name = stem + suffixFormat.format(i) + ext
        i += 1
        if stop >= 0 and i > stop:
            break

    return name


def englishTitleCase(text: str) -> str:
    if not text:
        return ""

    if QLocale().language() not in [QLocale.Language.C, QLocale.Language.English]:
        return text[0].upper() + text[1:]

    words = text.split()
    text = ""
    sep = ""
    for word in words:
        text += sep
        sep = " "
        if word in _titleLowercaseWords:
            text += word
        else:
            text += word[0].upper() + word[1:]

    return text


def naturalSort(text: str):
    text = text.casefold()
    parts = _naturalSortSplit.split(text)
    return [int(part) if part.isdigit() else part for part in parts]


def qstringLength(text: str) -> int:
    """ QString-compatible string length. """
    return len(text.encode("utf-16", "surrogatepass")) // 2 - 1



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.pycompat import *

import os
import re
import shlex
from enum import StrEnum
from pathlib import Path

from gitfourchette.localization import *
from gitfourchette.forms.textinputdialog import TextInputDialog
from gitfourchette.qt import *
from gitfourchette.sshagent import SshAgent
from gitfourchette.toolbox import escape, stockIcon, tquo, paragraphs, tweakWidgetFont, QHintButton

_ClearTextPromptPatterns = [
    # When connecting to an HTTPS remote with user/pass, the username is requested first.
    r"^Username(:| for )",

    # First time connecting to a host that's not in ~/.ssh/known_hosts
    r"Are you sure you want to continue connecting \(yes/no",

    # Follow-up question to the above
    r"Please type 'yes'"
]

_UnknownHostPrompt = "Are you sure you want to continue connecting (yes/no/[fingerprint])?"

_KeyFingerprintPattern = re.compile(r"key fingerprint is:? (.+?)\.?$", re.MULTILINE)


class AskpassPrompt(StrEnum):
    """
    AskpassDialog behaviors driven by the SSH_ASKPASS_PROMPT environment variable.
    """

    Entry = ""
    """
    Normal text input with ok/cancel buttons.
    """

    Confirm = "confirm"
    """
    No text input, yes/no buttons, exit code 0 if yes (e.g. `ssh -o AddKeysToAgent=ask`).
    Per `man ssh-add`: "Successful confirmation is signaled by a zero exit
    status from ssh-askpass, rather than text entered into the requester."
    """

    Message = "none"
    """
    No text input, just a dismiss button.
    """


class AskpassDialog(TextInputDialog):
    promptKind: AskpassPrompt | str

    autoYesOnAcceptEmptyText: bool
    "Reply 'yes' if the dialog is accepted with an empty text input."

    def __init__(self, parent: QWidget | None, prompt: str):
        promptKind = os.environ.get("SSH_ASKPASS_PROMPT", AskpassPrompt.Entry)
        self.promptKind = promptKind

        prompt = prompt.strip()
        clearText = any(re.search(pattern, prompt) for pattern in _ClearTextPromptPatterns)

        # Detect "unknown host" message
        self.unknownHostFingerprint = ""
        if clearText and promptKind == AskpassPrompt.Entry and prompt.endswith(_UnknownHostPrompt):
            fingerprintMatch = _KeyFingerprintPattern.search(prompt)
            if fingerprintMatch:
                self.unknownHostFingerprint = fingerprintMatch.group(1)
                prompt = prompt.removesuffix(_UnknownHostPrompt).strip()

        self.autoYesOnAcceptEmptyText = bool(self.unknownHostFingerprint)

        if promptKind == AskpassPrompt.Confirm:
            title = _("SSH is asking for your confirmation")
        elif promptKind == AskpassPrompt.Message:
            title = _("Message from SSH")
        elif self.unknownHostFingerprint:
            title = _("Connecting to unknown SSH host")
        else:
            title = _("Enter SSH credentials")

        subtitle = ""
        if promptKind == AskpassPrompt.Entry and not clearText:
            hasBuiltInAgent = bool(os.environ.get(SshAgent.EnvBuiltInAgentPid, ""))
            hasAgentSocket = os.path.exists(os.environ.get("SSH_AUTH_SOCK", ""))
            if hasBuiltInAgent:
                subtitle = _("{app}’s ssh-agent will remember this credential "
                             "until you quit the application.", app=qAppName())
            elif hasAgentSocket:
                subtitle = _("An ssh-agent is running on your system. It will remember this credential "
                             "if {0} is enabled in your SSH configuration.", tquo("AddKeysToAgent"))
            else:
                subtitle = _("This credential will not be remembered "
                             "because ssh-agent isn’t running on your system.")

        promptLines = escape(prompt).splitlines()
        if self.unknownHostFingerprint:
            promptLines.append("<b>" + _("To continue connecting, do you trust this key?"))
        htmlPrompt = paragraphs(*promptLines)

        super().__init__(
            parent, title, htmlPrompt, subtitle,
            multilineSubtitle=True, selectableLabel=True)

        if not clearText:
            self.lineEdit.setEchoMode(QLineEdit.EchoMode.Password)
            self.echoModeAction = self.lineEdit.addAction(stockIcon("view-visible"), QLineEdit.ActionPosition.TrailingPosition)
            self.echoModeAction.setToolTip(_("Reveal passphrase"))
            self.echoModeAction.triggered.connect(self.onToggleEchoMode)

        self.finished.connect(self.onFinish)

        if self.unknownHostFingerprint:
            self._setUpUnknownHostUi()
        elif promptKind == AskpassPrompt.Confirm:
            self.okButton.setText(_("Yes"))
            self.cancelButton.setText(_("No"))
            self.lineEdit.setVisible(False)
        elif promptKind == AskpassPrompt.Message:
            self.cancelButton.setVisible(False)
            self.lineEdit.setVisible(False)

    def _setUpUnknownHostUi(self):
        self.okButton.setText(_("Trust"))

        hintParts = [
            _("Check that the host’s key matches one you trust:"),
            "<ol><li><p>",
            _("Get the host’s fingerprint from a trusted source. "
              "(Do not use the one shown in this dialog!)"),
            "</p></li><li>",
            _("Paste the trusted fingerprint here. "
              "If it matches the host’s, you may trust this host and proceed."),
            "</p></li></ol>",
            _("This is safer than visually comparing fingerprints because it "
              "prevents “fuzzy fingerprint” attacks, where a malicious "
              "fingerprint is crafted to look like the real one."),
        ]

        checkbox = QCheckBox(_("&Verify fingerprint before trusting (optional)"))
        checkbox.setObjectName("VerifyFingerprintQCheckBox")
        checkbox.toggled.connect(self.onToggleFingerprintCheckbox)

        hint = QHintButton(toolTip="".join(hintParts))

        fingerprintWidget = QWidget()
        layout = QGridLayout(fingerprintWidget)
        layout.setContentsMargins(QMargins(0, 0, 0, 16))
        layout.addWidget(checkbox,      0, 0)
        layout.addWidget(hint,          0, 1)
        layout.addWidget(self.lineEdit, 1, 0, 1, 3)  # yank lineEdit out of its layout and into ours
        layout.setColumnStretch(2, 1)
        layout.setVerticalSpacing(0)
        tweakWidgetFont(fingerprintWidget, 80)

        self.setExtraWidget(fingerprintWidget)
        self.setValidator(self.validateFingerprint)

        # Prime unchecked state
        self.onToggleFingerprintCheckbox(True)
        self.onToggleFingerprintCheckbox(False)

    def onToggleEchoMode(self):
        passwordMode = self.lineEdit.echoMode() == QLineEdit.EchoMode.Password
        passwordMode = not passwordMode
        self.lineEdit.setEchoMode(QLineEdit.EchoMode.Password if passwordMode else QLineEdit.EchoMode.Normal)
        self.echoModeAction.setIcon(stockIcon("view-visible" if passwordMode else "view-hidden"))
        self.echoModeAction.setToolTip(_("Reveal passphrase") if passwordMode else _("Hide passphrase"))
        self.echoModeAction.setChecked(not passwordMode)

    def onToggleFingerprintCheckbox(self, checked: bool):
        self.lineEdit.setEnabled(checked)
        if checked:
            self.validator.successText = _("Both fingerprints match")
            self.lineEdit.setPlaceholderText(_("Enter fingerprint from trusted source"))
            if self.isVisible():
                self.lineEdit.setFocus()
        else:
            self.validator.successText = ""
            self.lineEdit.setPlaceholderText("")
            self.lineEdit.clear()
        self.validator.run(silenceEmptyWarnings=True)

    def validateFingerprint(self, s: str):
        if not self.lineEdit.isEnabled():
            return ""

        s = s.strip()
        if not s:
            return _("Cannot be empty.")
        if s != self.unknownHostFingerprint:
            return _("Fingerprint does not match.") + "\n" + self.unknownHostFingerprint
        return ""

    def onFinish(self, result: int):
        if not result:
            QApplication.instance().exit(1)
            return

        if self.promptKind == AskpassPrompt.Entry:
            secret = self.lineEdit.text()

            if not secret and self.autoYesOnAcceptEmptyText:
                secret = "yes"

            print(secret)

        QApplication.instance().exit(0)

    @classmethod
    def run(cls, prompt: str = ""):
        app = QApplication.instance()
        prompt = prompt or " ".join(app.arguments()[1:])
        dialog = AskpassDialog(None, prompt)
        dialog.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        dialog.show()
        return dialog

    @classmethod
    def environmentForChildProcess(cls, sandbox: bool):
        from gitfourchette.exttools.toolcommands import ToolCommands

        fileName = "askpass_sandboxed.sh" if sandbox else "askpass.sh"
        launcherScript = Path(qTempDir(), fileName)

        if not launcherScript.exists():
            tokens = ToolCommands.spawnNewInstance(__name__, sandbox=sandbox)

            # Discard stderr to avoid forwarding Qt error spam to ProcessDialog.
            script = ('#!/usr/bin/env bash\n'
                      f'exec {shlex.join(tokens)} "$@" 2>/dev/null\n')

            launcherScript.write_text(script, "utf-8")
            launcherScript.chmod(0o755)

        return {
            "SSH_ASKPASS": str(launcherScript),
            "SSH_ASKPASS_REQUIRE": "force",
        }


def main():
    from gitfourchette.application import GFApplication
    import sys
    app = GFApplication(sys.argv, barebones=True)
    _dontCollectMe = AskpassDialog.run()
    returnCode = app.exec()
    return sys.exit(returnCode)


if __name__ == "__main__":
    main()



# -----------------------------------------------------------------------------
# Copyright (C) 2025 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.ui_resetheaddialog import Ui_ResetHeadDialog
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *

DEFAULT_MODE = ResetMode.MIXED


class ResetHeadDialog(QDialog):
    activeMode: ResetMode

    def setActiveMode(self, mode: ResetMode):
        self.activeMode = mode
        okButton = self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        okButton.setIcon(self.defaultOkIcon)
        okButton.setToolTip("")
        if mode == ResetMode.HARD:
            okButton.setIcon(stockIcon("achtung"))
            okButton.setToolTip(_("Hard reset: Destructive action!"))
        self.ui.recurseCheckBox.setEnabled(self.hasSubmodules and mode == ResetMode.HARD)

    def recurseSubmodules(self):
        checkBox = self.ui.recurseCheckBox
        return checkBox.isEnabled() and checkBox.isChecked()

    def __init__(self, oid: Oid, branchName: str, commitText: str, hasSubmodules: bool, parent: QWidget):
        super().__init__(parent)

        self.ui = Ui_ResetHeadDialog()
        self.ui.setupUi(self)

        self.hasSubmodules = hasSubmodules
        self.ui.recurseCheckBox.setVisible(hasSubmodules)
        self.ui.recurseCheckBox.setEnabled(hasSubmodules)

        okButton = self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        self.defaultOkIcon = okButton.icon() or QIcon()

        self.modeButtons = {
            ResetMode.SOFT: self.ui.softButton,
            ResetMode.MIXED: self.ui.mixedButton,
            ResetMode.HARD: self.ui.hardButton,
        }
        self.modeLabels = {
            ResetMode.SOFT: self.ui.softHelp,
            ResetMode.MIXED: self.ui.mixedHelp,
            ResetMode.HARD: self.ui.hardHelp,
        }

        fontMetrics = self.modeButtons[ResetMode.SOFT].fontMetrics()
        spaceWidth = fontMetrics.horizontalAdvance(" " * 100) / 100
        desiredWidth = fontMetrics.horizontalAdvance("MixedWW")

        for mode, button in self.modeButtons.items():
            label = self.modeLabels[mode]
            formatWidgetText(label, commit=lquo(shortHash(oid)))
            tweakWidgetFont(label, 88)

            button.toggled.connect(lambda checked, m=mode: self.setActiveMode(m))

            # Pad checkbox text with spaces to enlarge clickable zone (hacky)
            missingWidth = desiredWidth - fontMetrics.horizontalAdvance(button.text())
            padding = " " * max(0, int(missingWidth / spaceWidth))
            button.setText(button.text() + padding)

        self.activeMode = DEFAULT_MODE
        self.modeButtons[DEFAULT_MODE].setChecked(True)

        title = _("Reset {0} to {1}", lquoe(branchName), lquo(shortHash(oid)))
        self.setWindowTitle(title)

        summary, _dummy = messageSummary(commitText)
        commitText = _("Commit {0}:", shortHash(oid)) + " " + tquo(summary)
        convertToBrandedDialog(self, subtitleText=commitText)

        packDialog(self, lockHeight=True)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""Non-textual diffs"""

from __future__ import annotations

import dataclasses
import os
import re
from contextlib import suppress
from pathlib import Path

from gitfourchette import settings
from gitfourchette.gitdriver import *
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator, NavContext, NavFlags
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.settings import WhitespaceMode
from gitfourchette.tasks import RepoTask
from gitfourchette.toolbox import *
from gitfourchette import trtables


# convert.c, since Git 2.37.0
_workdirCrlfWarning = re.compile(r"warning: in the working copy of .+, (.+) will be replaced by (.+) the next time Git touches it")
# Old message in versions older than 2.37.0
_workdirCrlfWarningLegacy = re.compile(r"warning: (.+) will be replaced by (.+) in .+\.")


class ImageDelta:
    @dataclasses.dataclass
    class ImageDeltaFile:
        size: int
        image: QImage | None
        deltaFile: GitDeltaFile

    old: ImageDeltaFile
    new: ImageDeltaFile

    def __init__(self, repo: Repo, delta: GitDelta):
        oldData = delta.old.read(repo)
        newData = delta.new.read(repo)

        self.old = ImageDelta.ImageDeltaFile(
            size=len(oldData),
            image=QImage.fromData(oldData) if oldData else None,
            deltaFile=delta.old)

        self.new = ImageDelta.ImageDeltaFile(
            size=len(newData),
            image=QImage.fromData(newData) if newData else None,
            deltaFile=delta.new)


class SpecialDiffError:
    def __init__(
            self,
            message: str,
            details: str = "",
            icon: str = "SP_MessageBoxInformation",
            preformatted: str = "",
            longform: str = "",
    ):
        self.message = message
        self.details = details
        self.icon = icon
        self.preformatted = preformatted
        self.longform = longform
        self.links = DocumentLinks()
        self.taskInvoker: QObject | None = None

    def taskLink(self, taskClass: type[RepoTask], *args, **kwargs) -> str:
        return self.links.new(lambda: taskClass.invoke(self.taskInvoker, *args, **kwargs))

    @staticmethod
    def noChange(repo: Repo, delta: GitDelta, stderr: str = ""):
        message = _("File contents didn’t change.")
        details: list[str] = []
        longform: list[str] = []

        oldFile = delta.old
        newFile = delta.new
        oldFileExists = not oldFile.isId0()
        newFileExists = not newFile.isId0()

        if not newFileExists:
            message = _("Empty file was deleted.")

        if not oldFileExists:
            if delta.new.mode == FileMode.TREE:
                return SpecialDiffError.treeDiff(repo, delta)
            else:
                assert delta.status.isAddedOrUntracked
                message = _("New empty file.")

        if (newFileExists
                and oldFileExists
                and newFile.id != oldFile.id
                and settings.prefs.whitespaceMode != WhitespaceMode.Strict):
            message = _("Whitespace changes ignored. Contents otherwise identical.")
            detailsLine = "{} {}.".format(_("Current whitespace mode:"), trtables.enum(settings.prefs.whitespaceMode))
            details.append(detailsLine)

        if oldFile.path != newFile.path:
            intro = _("Renamed:")
            details.append(f"{intro} {hquo(oldFile.path)} &rarr; {hquo(newFile.path)}.")

        if oldFileExists and oldFile.mode != newFile.mode:
            intro = _("Mode change:")
            details.append(f"{intro} {trtables.enum(oldFile.mode)} &rarr; {trtables.enum(newFile.mode)}.")

        if delta.source == GitDeltaSource.Dirty and stderr:
            crlfWarningPattern = _workdirCrlfWarning
            if GitDriver.gitVersionTuple() < (2, 37):
                crlfWarningPattern = _workdirCrlfWarningLegacy
            crlfWarningMatch = crlfWarningPattern.search(stderr)
            if crlfWarningMatch:
                sourceEndings = crlfWarningMatch.group(1)
                targetEndings = crlfWarningMatch.group(2)
                longform.append(toRoomyUL([
                    _("In your working copy, {0} will be replaced by {1} the next time Git touches this file.", sourceEndings, targetEndings),
                    _("You can stage the file to dismiss this message; no changes will be recorded.")
                ]))

        return SpecialDiffError(message, "\n".join(details), longform="\n".join(longform))

    @staticmethod
    def fileTooLarge(size: int, threshold: int, locator: NavLocator, image: bool, legacy: bool = False):
        if image:
            prefKey = "imageFileThresholdKB"
            titleText = _("This image is very large.")
            loadText = _("[Load image anyway] (this may take a moment)")
            configText = _("[Configure image preview limit] (currently: {0})")
        else:
            prefKey = "largeFileThresholdKB"
            titleText = _("This file is very large.") if not legacy else _("This diff is very large.")
            loadText = _("[Load diff anyway] (this may take a moment)")
            configText = _("[Configure diff preview limit] (currently: {0})")

        locale = QLocale()
        humanSize = locale.formattedDataSize(size, 1) if not legacy else ""  # legacy: hide size
        humanThreshold = locale.formattedDataSize(threshold, 0)
        loadLink = locator.withExtraFlags(NavFlags.AllowLargeFiles).url()
        configLink = makeInternalLink("prefs", prefKey)

        longform = toRoomyUL([
            linkify(loadText, loadLink),
            linkify(configText.format(humanThreshold), configLink),
        ])

        return SpecialDiffError(titleText, humanSize, "SP_MessageBoxWarning", longform=longform)

    @staticmethod
    def typeChange(delta: GitDelta):
        oldText = _("Old type:")
        newText = _("New type:")
        oldMode = trtables.enum(delta.old.mode)
        newMode = trtables.enum(delta.new.mode)
        table = ("<table>"
                 f"<tr><td><del><b>{oldText}</b></del> </td><td>{oldMode}</tr>"
                 f"<tr><td><add><b>{newText}</b></add> </td><td>{newMode}</td></tr>"
                 "</table>")
        return SpecialDiffError(_("This file’s type has changed."), table)

    @staticmethod
    def binaryDiff(repo: Repo, delta: GitDelta, locator: NavLocator) -> SpecialDiffError | ImageDelta:
        locale = QLocale()
        oldSize = delta.old.sizeBallpark(repo)
        newSize = delta.new.sizeBallpark(repo)

        if isImageFormatSupported(delta.old.path) and isImageFormatSupported(delta.new.path):
            largestSize = max(oldSize, newSize)
            if locator.hasFlags(NavFlags.AllowLargeFiles):
                threshold = 0
            else:
                threshold = settings.prefs.imageFileThresholdKB * 1024
            if largestSize > threshold > 0:
                return SpecialDiffError.fileTooLarge(largestSize, threshold, locator, image=True)
            return ImageDelta(repo, delta)

        oldSizeText = locale.formattedDataSize(oldSize)
        newSizeText = locale.formattedDataSize(newSize)
        sizeText = f"{oldSizeText} &rarr; {newSizeText}"
        return SpecialDiffError(_("File appears to be binary."), sizeText)

    @staticmethod
    def missingLfsObjects(error: LfsObjectCacheMissingError) -> SpecialDiffError:
        from gitfourchette.tasks import DownloadLfsObjects

        locale = QLocale()
        numObjects = len(error.pointers)
        title = _n("Object missing from local LFS cache.",
                   "Objects missing from local LFS cache.", n=numObjects)
        specialDiff = SpecialDiffError(title, icon="SP_MessageBoxWarning")

        totalMissingSize = sum(ptr.size for ptr in error.pointers)
        humanMissingSize = locale.formattedDataSize(totalMissingSize)

        loadUrl = specialDiff.taskLink(DownloadLfsObjects, error)
        loadText = _n("Download object to display the diff",
                      "Download objects to display the diff", n=numObjects)
        specialDiff.details = f"{linkify(loadText, loadUrl)} ({humanMissingSize})"

        specialDiff.longform = "<small>" + "<br>".join(
            f"LFS oid {hquo(shortHash(p.id))}, {locale.formattedDataSize(p.size)}"
            for p in error.pointers)

        return specialDiff

    @staticmethod
    def lfsMigration(repo: Repo, delta: GitDelta):
        flagNames = [englishTitleCase(_("not LFS")), "LFS"]

        hasOldLfs = bool(delta.old.lfs)
        hasNewLfs = bool(delta.new.lfs)
        lfsSide = delta.new if hasNewLfs else delta.old
        blobSide = delta.old if hasNewLfs else delta.new
        assert hasOldLfs != hasNewLfs

        title = " ".join([
            _("Storage changed:"),
            f"<b>{flagNames[hasOldLfs]}</b> &rarr; <b>{flagNames[hasNewLfs]}</b>",
            stockIconImgTag("git-lfs-add" if hasNewLfs else "git-lfs-remove"),
        ])
        details = ""

        if not lfsSide.lfs.isTentative():
            blobSha256 = blobSide.blobSha256(repo)
            if blobSha256 == lfsSide.lfs.id:
                details = _("File contents remained identical during the conversion.")
            else:
                icon = stockIconImgTag("achtung") + " "
                details = icon + _("File contents modified during the conversion!")

        return SpecialDiffError(title, details)

    @staticmethod
    def treeDiff(repo: Repo, delta: GitDelta):
        from gitfourchette.tasks import AbsorbSubmodule

        message = _("This untracked subtree is the root of another Git repository.")
        specialDiff = SpecialDiffError(message)

        treePath = os.path.normpath(delta.new.path)
        treeName = os.path.basename(treePath)

        prompt1 = _("Open {0}", bquo(treeName))
        openLink = QUrl.fromLocalFile(repo.in_workdir(treePath)).toString()

        prompt2 = _("Absorb {0} as submodule", bquo(treeName))
        prompt2 = _("Recommended action:") + " [" + prompt2 + "]"
        absorbLink = specialDiff.taskLink(AbsorbSubmodule, path=treePath)

        specialDiff.details = linkify(prompt1, openLink)
        specialDiff.longform = toRoomyUL([linkify(prompt2, absorbLink)])
        return specialDiff

    @staticmethod
    def submoduleDiff(repo: Repo, delta: GitDelta, locator: NavLocator) -> SpecialDiffError:
        from gitfourchette.tasks import AbsorbSubmodule, DiscardFiles, RegisterSubmodule

        assert locator.context == NavContext.fromGitDeltaSource(delta.source)

        path = delta.new.path
        absPath = repo.in_workdir(path)
        stillExists = Path(absPath).is_dir()
        isAbsorbed = stillExists and Path(absPath, ".git").is_file()

        isDel = delta.status == GitStatus.Deleted
        isAdd = delta.status.isAddedOrUntracked
        oldId = delta.old.id
        newId = delta.new.id
        headDidMove = oldId != newId

        try:
            name = next(n for n, p in repo.listall_submodules_dict().items() if p == path)
            isRegistered = True
        except StopIteration:
            # Name may not be available when viewing a past commit about a deleted submodule,
            # or a tree that isn't registered as a submodule yet
            name = path  # Fallback to inner path as the name
            isRegistered = False

        if delta.source.isWorkdir():
            wasRegistered = path in repo.listall_submodules_dict_at_head()
        else:
            wasRegistered = True

        isSubmodule = isRegistered or wasRegistered
        isTree = not isSubmodule

        # Compose title.
        # Explicit permutations of "subtree"/"submodule" text so that translations
        # can be grammatically correct (in case of different genders, etc.)
        if isDel:
            title = _("Subtree {0} was [removed.]") if isTree else _("Submodule {0} was [removed.]")
            title = tagify(title, "<del><b>")
        elif isAdd:
            title = _("Subtree {0} was [added.]") if isTree else _("Submodule {0} was [added.]")
            title = tagify(title, "<add><b>")
        elif headDidMove:
            title = _("Subtree {0} was updated.") if isTree else _("Submodule {0} was updated.")
        else:
            title = _("Subtree {0} contains changes.") if isTree else _("Submodule {0} contains changes.")

        title = title.format(bquo(name))

        # Add link to open the submodule as a subtitle
        subtitle = ""
        openLink = QUrl.fromLocalFile(absPath)
        if stillExists:
            subtitle = _("Open subtree") if isTree else _("Open submodule")
            if name != path:
                subtitle += " " + _("(path: {0})", escape(path))
            subtitle = linkify(subtitle, openLink)

        # Initialize SpecialDiffError (we'll return this)
        specialDiff = SpecialDiffError(title, subtitle)
        longformParts = []

        # Create old/new table if the submodule's HEAD commit was moved
        if headDidMove and not isDel:
            targets = [shortHash(oldId), shortHash(newId)]
            messages = ["", ""]

            # Show additional details about the commits if there's still a workdir for this submo
            if stillExists:
                try:
                    with RepoContext(absPath, RepositoryOpenFlag.NO_SEARCH) as subRepo:
                        for i, h in enumerate([oldId, newId]):
                            if h == NULL_OID:
                                continue

                            # Link to specific commit
                            targets[i] = linkify(shortHash(h), f"{openLink.toString()}#{h}")

                            # Get commit summary
                            with suppress(LookupError, GitError):
                                m = subRepo[h].peel(Commit).message
                                m = messageSummary(m)[0]
                                m = elide(m, Qt.TextElideMode.ElideRight, 25)
                                m = hquo(m)
                                messages[i] = m
                except GitError:
                    # RepoContext may fail if the submodule couldn't be opened for any reason.
                    # Don't show an error for this, show the diff document anyway
                    pass

            intro = (_("The subtree’s <b>HEAD</b> has moved to another commit.") if isTree else
                     _("The submodule’s <b>HEAD</b> has moved to another commit."))
            if delta.source == GitDeltaSource.Dirty:
                intro += " " + _("You can stage this update:")
            longformParts.append(
                f"{intro}"
                "<p><table>"
                f"<tr><td><del><b>{_('Old:')}</b></del> </td><td><tt>{targets[0]}</tt> {messages[0]}</td></tr>"
                f"<tr><td><add><b>{_('New:')}</b></add> </td><td><tt>{targets[1]}</tt> {messages[1]}</td></tr>"
                "</table></p>")

        # Show additional tips if this submodule is in the workdir.
        if delta.source.isWorkdir():
            m = ""
            if isDel:
                if isRegistered:
                    m = _("To complete the removal of this submodule, <b>remove it from {gitmodules}</b>.")
                elif wasRegistered:
                    m = _("To complete the removal of this submodule, make sure to <b>commit "
                          "{gitmodules}</b> at the same time as the submodule folder itself.")

            elif isRegistered and not wasRegistered:
                m = _("To complete the addition of this submodule, make sure to <b>commit "
                      "{gitmodules}</b> at the same time as the submodule folder itself.")

            elif not isAbsorbed:
                if isTree:
                    m = _("<b>This subtree isn’t a submodule yet!</b> "
                          "You should [absorb this subtree] into the parent repository so it becomes a submodule.")
                else:
                    m = _("To complete the addition of this submodule, "
                          "you should [absorb the submodule] into the parent repository.")
                m = linkify(m, specialDiff.taskLink(AbsorbSubmodule, path=path))

            elif not isRegistered:
                m = _("To complete the addition of this submodule, [register it in {gitmodules}].")
                m = linkify(m, specialDiff.taskLink(RegisterSubmodule, path=path))

            if m:
                m = m.format(gitmodules=f"<tt>{DOT_GITMODULES}</tt>")
                m = f"{stockIconImgTag('achtung')} <b>{_('IMPORTANT')}</b> &ndash; {m}"
                longformParts.insert(0, m)

            # Tell about any uncommitted changes
            if delta.submoduleWorkdirDirty:
                discardLink = specialDiff.taskLink(DiscardFiles, [delta])

                if isTree:
                    uc1 = _("The subtree contains <b>uncommitted changes</b>. They can’t be committed from the parent repo. You can:")
                    uc2 = _("[Open] the subtree and commit the changes.")
                    uc3 = _("Or, [Reset] the subtree to a clean state.")
                else:
                    uc1 = _("The submodule has <b>uncommitted changes</b>. They can’t be committed from the parent repo. You can:")
                    uc2 = _("[Open] the submodule and commit the changes.")
                    uc3 = _("Or, [Reset] the submodule to a clean state.")

                m = f"{uc1}<ul><li>{uc2}</li><li>{uc3}</li></ul>"
                m = linkify(m, openLink, discardLink)
                longformParts.append(m)

        # Compile longform parts into an unordered list
        specialDiff.longform = toRoomyUL(longformParts)

        return specialDiff



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *


class CodeRubberBand(QWidget):
    def paintEvent(self, event: QPaintEvent):
        # Don't inherit QRubberBand (pen thickness ignored on Linux!)
        RX = 12  # rounded rect x radius
        RY = 4  # rounded rect y radius
        T = 4  # thickness
        HT = T//2  # half thickness
        CT = T//2  # clipped thickness
        OT = 1  # outline extra thickness

        palette = self.palette()
        painter = QPainter(self)
        rect: QRect = self.rect().marginsRemoved(QMargins(HT, HT, HT, HT))

        outlineColor: QColor = palette.color(QPalette.ColorRole.Base)
        outlineColor.setAlphaF(.75 if outlineColor.lightnessF() < .5 else .5)  # light mode: subtler alpha

        if self.parentWidget().hasFocus():
            penColor = palette.accent().color()
        else:
            penColor = palette.color(QPalette.ColorGroup.Inactive, QPalette.ColorRole.Highlight)

        pen = QPen(penColor, T)
        outlinePen = QPen(outlineColor, T+2*OT)

        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        painter.setClipRect(rect.marginsAdded(QMargins(HT, CT-HT, HT, CT-HT)))

        path = QPainterPath()
        path.addRoundedRect(QRectF(rect), RX, RY, Qt.SizeMode.AbsoluteSize)

        painter.setClipRect(rect.marginsAdded(QMargins(HT, CT-HT+OT, HT, CT-HT+OT)))
        painter.setPen(outlinePen)
        painter.drawPath(path)

        painter.setClipRect(rect.marginsAdded(QMargins(HT, CT-HT, HT, CT-HT)))
        painter.setPen(pen)
        painter.drawPath(path)

        painter.end()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging
import os.path
import re
from typing import ClassVar

import pygments.lexers
from pygments.lexer import Lexer
from pygments import __version__ as pygmentsVersion

from gitfourchette.toolbox.benchmark import benchmark

logger = logging.getLogger(__name__)

# Override Pygments priorities
_disambiguations = {
    ".ASM": "nasm",
    ".G": None,
    ".S": "gas",
    ".as": "actionscript3",
    ".asax": "aspx-cs",
    ".ascx": "aspx-cs",
    ".ashx": "aspx-cs",
    ".asm": "nasm",
    ".asmx": "aspx-cs",
    ".aspx": "aspx-cs",
    ".axd": "aspx-cs",
    ".bas": "qbasic",
    ".cl": "common-lisp",
    ".cp": "cpp",
    ".ecl": "prolog",
    ".fs": "fsharp",
    ".g": None,
    ".gd": "gdscript",
    ".h": "c",
    ".hh": "cpp",
    ".html": "html",
    ".inc": "php",
    ".inf": "ini",
    ".m": "objective-c",
    ".mm": "objective-c++",
    ".pl": "perl6",
    ".pm": "perl6",
    ".pro": "prolog",
    ".prolog": "prolog",
    ".rl": None,
    ".s": "gas",
    ".sql": "sql",
    ".t": "perl6",
    ".toc": "tex",
    ".tst": "scilab",
    ".ttl": "turtle",
    ".v": "verilog",
    ".xml": "xml",
    ".xsl": "xslt",
    ".xslt": "xslt",
}

# These are not defined by Pygments
_forcedAssociations = {
    ".gitmodules": "ini",
    ".svg": "xml",
}


class LexerCache:
    """
    Fast drop-in replacement for pygments.lexers.get_lexer_for_filename().
    """

    lexerAliases: ClassVar[dict[str, str]] = {}
    " Lexer aliases by file extensions or verbatim file names "

    lexerInstances: ClassVar[dict[str, Lexer]] = {}
    " Lexer instances by aliases "

    @classmethod
    @benchmark
    def getLexerFromPath(cls, path: str, allowPlugins: bool) -> Lexer | None:
        assert path

        if not cls.lexerAliases:
            cls.warmUp(allowPlugins)

        # Find lexer alias by verbatim filename or extension
        try:
            # Try verbatim name (e.g. 'CMakeLists.txt')
            fileName = os.path.basename(path)
            alias = cls.lexerAliases[fileName]
        except KeyError:
            # Find lexer alias by extension
            _dummy, ext = os.path.splitext(path)
            alias = cls.lexerAliases.get(ext, "")

        # Bail early
        if not alias:
            return None

        # Get existing lexer instance
        instance = cls.lexerInstances.get(alias)
        if instance is not None:
            return instance

        # Instantiate new lexer.
        # Notes:
        # - Passing in an alias from pygments' builtin lexers shouldn't
        #   tap into Pygments plugins, so this should be fairly fast.
        # - stripnl throws off highlighting in files that begin with
        #   whitespace.
        lexer = pygments.lexers.get_lexer_by_name(alias, stripnl=False)
        cls.lexerInstances[alias] = lexer
        return lexer

    @classmethod
    @benchmark
    def warmUp(cls, allowPlugins: bool) -> None:
        """
        Cache lexerAliases (map of filename patterns to lexer names).
        Turn off plugins for a significant speedup.
        """

        # Map lexer names to filename patterns
        lexers: dict[str, set[str]] = {}

        simpleExtension = re.compile(r'^\*\.[^.*\[]+$')
        for _name, aliases, patterns, _mimeTypes in pygments.lexers.get_all_lexers(plugins=allowPlugins):
            if not patterns or not aliases:  # Skip lexers without filename associations
                continue

            lexerName = aliases[0]
            patternSet: set[str] = set()
            if lexerName in lexers:  # pragma: no cover
                logger.warning(f"Duplicated lexer name: {lexerName}")
            lexers[lexerName] = patternSet

            for pattern in patterns:
                if simpleExtension.match(pattern):  # *.ext
                    pattern = pattern[1:]  # Strip "*" prefix
                elif '*' in pattern:
                    # Skip anything else with a wildcard (keep verbatim filenames)
                    continue

                if lexerName == _disambiguations.get(pattern, lexerName):
                    patternSet.add(pattern)

        # Patch missing extensions
        for ext, lexerName in _forcedAssociations.items():
            try:
                lexers[lexerName].add(ext)
            except KeyError:  # pragma: no cover
                logger.warning(f"Missing lexer {lexerName}")

        # Avoid Pygments overhead for plaintext lexers
        lexers.pop("text", None)

        # Disable Lua syntax highlighting in bad Pygments versions (issue #55)
        if pygmentsVersion.startswith(("2.19.0", "2.19.1")):  # pragma: no cover
            logger.warning(f"Disabling Lua support in Pygments {pygmentsVersion} (issue #55)")
            lexers.pop("lua", None)

        # Map filename patterns to lexers
        patternsToLexers = {}
        for lexerName, patternSet in lexers.items():
            if False:  # pragma: no cover - replace with 'if True' to debug missing disambiguations
                overwrite = patternSet.intersection(patternsToLexers)  # type: ignore[unreachable]
                if overwrite:
                    logger.warning(f"Missing disambiguation: Lexer '{lexerName}' will take precedence over { { p: patternsToLexers[p] for p in overwrite } }")

            patternsToLexers.update(dict.fromkeys(patternSet, lexerName))

        cls.lexerAliases = patternsToLexers



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.signatureform import SignatureForm
from gitfourchette.forms.ui_identitydialog import Ui_IdentityDialog
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class IdentityDialog(QDialog):
    def __init__(
            self,
            firstRun: bool,
            initialName: str,
            initialEmail: str,
            configPath: str,
            repoHasLocalIdentity: bool,
            parent: QWidget
    ):
        super().__init__(parent)

        ui = Ui_IdentityDialog()
        ui.setupUi(self)
        self.ui = ui

        formatWidgetText(ui.configPathLabel, lquo(compactPath(configPath)))
        ui.warningLabel.setVisible(repoHasLocalIdentity)

        # Initialize with global identity values (if any)
        ui.nameEdit.setText(initialName)
        ui.emailEdit.setText(initialEmail)

        validator = ValidatorMultiplexer(self)
        validator.setGatedWidgets(ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok))
        validator.connectInput(ui.nameEdit, SignatureForm.validateInput)
        validator.connectInput(ui.emailEdit, SignatureForm.validateInput)
        validator.run(silenceEmptyWarnings=True)

        subtitle = _("This information will be embedded in the commits and tags that you create on this machine.")
        if firstRun:
            subtitle = _("Before editing this repository, please set up your identity for Git.") + " " + subtitle

        convertToBrandedDialog(self, subtitleText=subtitle, multilineSubtitle=True)
        packDialog(self, lockHeight=True)

    def identity(self) -> tuple[str, str]:
        name = self.ui.nameEdit.text()
        email = self.ui.emailEdit.text()
        return name, email



# Form implementation generated from reading ui file 'stashdialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_StashDialog(object):
    def setupUi(self, StashDialog):
        StashDialog.setObjectName("StashDialog")
        StashDialog.setWindowModality(Qt.WindowModality.NonModal)
        StashDialog.setEnabled(True)
        StashDialog.resize(461, 432)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(StashDialog.sizePolicy().hasHeightForWidth())
        StashDialog.setSizePolicy(sizePolicy)
        StashDialog.setSizeGripEnabled(False)
        StashDialog.setModal(True)
        self.vboxlayout = QVBoxLayout(StashDialog)
        self.vboxlayout.setObjectName("vboxlayout")
        self.messageEdit = QLineEdit(parent=StashDialog)
        self.messageEdit.setObjectName("messageEdit")
        self.vboxlayout.addWidget(self.messageEdit)
        self.keepCheckBox = QCheckBox(parent=StashDialog)
        self.keepCheckBox.setChecked(False)
        self.keepCheckBox.setObjectName("keepCheckBox")
        self.vboxlayout.addWidget(self.keepCheckBox)
        spacerItem = QSpacerItem(20, 16, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.vboxlayout.addItem(spacerItem)
        self.willStashLabel = QLabel(parent=StashDialog)
        self.willStashLabel.setText("(&WILL STASH THESE:)")
        self.willStashLabel.setWordWrap(True)
        self.willStashLabel.setObjectName("willStashLabel")
        self.vboxlayout.addWidget(self.willStashLabel)
        self.fileList = QListWidget(parent=StashDialog)
        self.fileList.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.fileList.setIconSize(QSize(16, 16))
        self.fileList.setObjectName("fileList")
        self.vboxlayout.addWidget(self.fileList)
        self.indexAndWtWarning = QLabel(parent=StashDialog)
        self.indexAndWtWarning.setWordWrap(True)
        self.indexAndWtWarning.setObjectName("indexAndWtWarning")
        self.vboxlayout.addWidget(self.indexAndWtWarning)
        self.buttonBox = QDialogButtonBox(parent=StashDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setCenterButtons(False)
        self.buttonBox.setObjectName("buttonBox")
        self.vboxlayout.addWidget(self.buttonBox)
        self.willStashLabel.setBuddy(self.fileList)

        self.retranslateUi(StashDialog)
        self.buttonBox.rejected.connect(StashDialog.reject)
        self.buttonBox.accepted.connect(StashDialog.accept)

    def retranslateUi(self, StashDialog):
        StashDialog.setWindowTitle(_p("StashDialog", "New stash"))
        self.messageEdit.setPlaceholderText(_p("StashDialog", "Optional stash message"))
        self.keepCheckBox.setToolTip(_p("StashDialog", "Normally, stashed changes are removed from the working directory. Tick this to leave the stashed changes intact instead."))
        self.keepCheckBox.setText(_p("StashDialog", "&Retain stashed changes in working directory"))
        self.indexAndWtWarning.setText(_p("StashDialog", "Warning: Some of the files have both staged and unstaged changes. Those changes will be combined in the stash."))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging
import re

from gitfourchette.exttools.toolcommands import ToolCommands
from gitfourchette.qt import *

logger = logging.getLogger(__name__)


class SshAgent(QProcess):
    EnvBuiltInAgentPid = "GITFOURCHETTE_AGENT"

    environment: dict[str, str]
    sandboxed: bool

    def __init__(self, parent: QObject, sandbox: bool = False):
        super().__init__(parent)
        self.setObjectName("SshAgent")

        program = "ssh-agent"

        if FLATPAK and sandbox:
            program = ToolCommands.FlatpakSandboxedCommandPrefix + program
            self.sandboxed = True
        else:
            self.sandboxed = False

        tokens = [program, "-c", "-D"]
        process = self
        process.setProgram(tokens[0])
        process.setArguments(tokens[1:])
        ToolCommands.wrapFlatpakCommand(process)
        process.start()
        if not process.waitForStarted():
            raise RuntimeError("ssh-agent did not start")
        process.waitForReadyRead()

        output = process.readAll().data().decode("utf-8", errors="replace")

        pidMatch = re.search(r"Agent pid (.+);", output, re.IGNORECASE)
        sockMatch = re.search(r"setenv SSH_AUTH_SOCK (.+);", output)
        if not sockMatch or not pidMatch:
            raise ValueError("didn't find SSH_AUTH_SOCK or agent PID")

        # It's tempting to get the PID via QProcess instead of parsing stdout,
        # but remember that we may be launching ssh-agent via flatpak-spawn
        # (which has a separate PID).
        sshAgentPid = pidMatch.group(1)
        sshAuthSock = sockMatch.group(1)

        self.environment = {
            "SSH_AUTH_SOCK": sshAuthSock,
            "SSH_AGENT_PID": sshAgentPid,
            SshAgent.EnvBuiltInAgentPid: sshAgentPid,
        }

        logger.info(f"ssh-agent started on PID {sshAgentPid} ({sshAuthSock})")

    def isSandboxed(self) -> bool:
        return self.sandboxed

    def stopAndWait(self, msec: int = 500):
        if self.state() != QProcess.ProcessState.Running:
            return

        for stop, name in (self.terminate, "SIGTERM"), (self.kill, "SIGKILL"):
            stop()
            if self.waitForFinished(msec):
                logger.info(f"ssh-agent stopped ({name})")
                break
            logger.warning(f"ssh-agent didn't respond to {name} after {msec} ms")



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Qt objects should always be collected on the main thread. However, Python's
automatic garbage collector may kick in at any time, including on a background
thread - possibly causing Qt to crash.

Import this module to turn off automatic GC and call gcHint() to collect the
garbage at opportune times.
"""

import gc
import logging

from gitfourchette.appconsts import APP_DEBUG, APP_VERBOSEDEL
from gitfourchette.qt import QApplication, QThread

_logger = logging.getLogger(__name__)

gc.disable()


def _onAppThread():
    app = QApplication.instance()
    return (app is None) or (app.thread() is QThread.currentThread())


if APP_DEBUG:
    def checkNoGcOnBackgroundThreads(phase, _info):
        if phase == 'start' and not _onAppThread():
            _logger.warning("!!!DANGER!!! Garbage collection on non-UI thread!")

    gc.callbacks.append(checkNoGcOnBackgroundThreads)


def gcHint():
    assert _onAppThread(), "garbage collection should only occur on the UI thread"

    for gen, c, t in zip(range(3), gc.get_count(), gc.get_threshold(), strict=True):
        if c <= t:
            if APP_VERBOSEDEL:
                _logger.debug(f"GC gen {gen}: {c}/{t}, stop")
            break
        num = gc.collect(gen)
        if APP_VERBOSEDEL:
            _logger.debug(f"GC gen {gen}: {c}/{t}, {num}")



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging
from collections.abc import Iterable
from contextlib import suppress
from typing import Any, overload, ClassVar

from gitfourchette import trtables
from gitfourchette import settings
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import RepoModel, UC_FAKEREF
from gitfourchette.repoprefs import RefSort
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)

BRANCH_FOLDERS = True
SYMBOL_AHEAD = "\u2191"
SYMBOL_BEHIND = "\u2193"
SYMBOL_UPDOWN = "\u21c5"


class SidebarItem(enum.IntEnum):
    Root = -1
    Spacer = 0
    WorkdirHeader = enum.auto()
    UncommittedChanges = enum.auto()
    LocalBranchesHeader = enum.auto()
    StashesHeader = enum.auto()
    RemotesHeader = enum.auto()
    TagsHeader = enum.auto()
    SubmodulesHeader = enum.auto()
    LocalBranch = enum.auto()
    DetachedHead = enum.auto()
    UnbornHead = enum.auto()
    Stash = enum.auto()
    Remote = enum.auto()
    RemoteBranch = enum.auto()
    Tag = enum.auto()
    Submodule = enum.auto()
    RefFolder = enum.auto()


class SidebarLayout:
    RootItems: ClassVar = [
        SidebarItem.WorkdirHeader,
        SidebarItem.UncommittedChanges,
        SidebarItem.Spacer,
        SidebarItem.LocalBranchesHeader,
        SidebarItem.Spacer,
        SidebarItem.RemotesHeader,
        SidebarItem.Spacer,
        SidebarItem.TagsHeader,
        SidebarItem.Spacer,
        SidebarItem.StashesHeader,
        SidebarItem.Spacer,
        SidebarItem.SubmodulesHeader,
    ]

    ForceExpand: ClassVar = [
        SidebarItem.WorkdirHeader
    ]

    NonleafItems: ClassVar = sorted([
        SidebarItem.Root,
        SidebarItem.WorkdirHeader,
        SidebarItem.LocalBranchesHeader,
        SidebarItem.RefFolder,
        SidebarItem.Remote,
        SidebarItem.RemotesHeader,
        SidebarItem.StashesHeader,
        SidebarItem.SubmodulesHeader,
        SidebarItem.TagsHeader,
    ])

    UnindentItems: ClassVar = {
        SidebarItem.LocalBranch: -1,
        SidebarItem.UnbornHead: -1,
        SidebarItem.DetachedHead: -1,
        SidebarItem.Stash: -1,
        SidebarItem.Tag: -1,
        SidebarItem.Submodule: -1,
        SidebarItem.Remote: -1,
        SidebarItem.RemoteBranch: -1,
        SidebarItem.RefFolder: -1,
    }

    HideableItems: ClassVar = sorted([
        SidebarItem.LocalBranch,
        SidebarItem.Remote,
        SidebarItem.RemoteBranch,
        SidebarItem.RefFolder,
    ])


class SidebarNode:
    children: list[SidebarNode]
    parent: SidebarNode | None
    row: int
    kind: SidebarItem
    data: str
    warning: str
    displayName: str

    @staticmethod
    def fromIndex(index: QModelIndex) -> SidebarNode:
        if not index.isValid():
            raise NotImplementedError("Can't make a SidebarNode from an invalid QModelIndex!")
        p = index.internalPointer()
        assert isinstance(p, SidebarNode)
        return p

    def __init__(self, kind: SidebarItem, data: str = ""):
        self.children = []
        self.parent = None
        self.row = -1
        self.kind = kind
        self.data = data
        self.warning = ""
        self.displayName = ""

    def appendChild(self, node: SidebarNode):
        assert self.mayHaveChildren()
        assert self is not node
        assert not node.parent
        node.row = len(self.children)
        node.parent = self
        self.children.append(node)

    def findChild(self, kind: SidebarItem, data: str = "") -> SidebarNode:
        """ Warning: this is inefficient - don't use this if there are many children! """
        assert self.mayHaveChildren()
        with suppress(StopIteration):
            return next(c for c in self.children if c.kind == kind and c.data == data)
        raise KeyError("child node not found")

    def getCollapseHash(self) -> str:
        assert self.mayHaveChildren(), "it's futile to hash a leaf"
        return f"{self.kind.name}.{self.data}"
        # Warning: it's tempting to replace this with something like "hash(data) << 8 | item",
        # but hash(data) doesn't return stable values across different Python sessions,
        # so it's not suitable for persistent storage (in history.json).

    def mayHaveChildren(self):
        return self.kind in SidebarLayout.NonleafItems

    def wantForceExpand(self):
        return self.kind in SidebarLayout.ForceExpand

    def canBeHidden(self):
        return self.kind in SidebarLayout.HideableItems

    def walk(self):
        # Unit test helper
        frontier = self.children[:]
        while frontier:
            node = frontier.pop(0)
            yield node
            frontier.extend(node.children)

    def isSimilarEnoughTo(self, other: SidebarNode):
        """ Use this to compare SidebarNodes from two different models. """
        return self.kind == other.kind and self.data == other.data

    def isLeafBranchKind(self):
        return self.kind == SidebarItem.LocalBranch or self.kind == SidebarItem.RemoteBranch

    def refMatchingPattern(self):
        if self.isLeafBranchKind():
            return self.data

        if self.kind == SidebarItem.Remote:
            return f"{RefPrefix.REMOTES}{self.data}/"

        if self.kind == SidebarItem.RefFolder:
            return f"{self.data}/"

        return ""

    def __repr__(self):
        return f"SidebarNode({self.kind.name} {self.data})"


class SidebarModel(QAbstractItemModel):
    repoModel: RepoModel
    rootNode: SidebarNode
    nodesByRef: dict[str, SidebarNode]
    _unbornHead: str

    _checkedOut: str
    "Shorthand of checked-out local branch"

    _checkedOutUpstream: str
    "Shorthand of the checked-out branch's upstream"

    _cachedToolTipIndex: QModelIndex
    _cachedToolTipText: str

    collapseCacheLayers: list[set[str]]
    """
    Keeps a cache of collapsed nodes.

    Layer #0 is the permanent state. It is saved to disk as part of RepoPrefs.
    It is applied when SidebarFilter is inactive.

    Layer #1 is the transient state. It only exists as long as SidebarFilter is
    active.
    """

    class Role:
        Ref = Qt.ItemDataRole(Qt.ItemDataRole.UserRole + 0)
        IconKey = Qt.ItemDataRole(Qt.ItemDataRole.UserRole + 1)
        AheadBehind = Qt.ItemDataRole(Qt.ItemDataRole.UserRole + 2)
        MissingUpstream = Qt.ItemDataRole(Qt.ItemDataRole.UserRole + 3)

    @property
    def _parentWidget(self) -> QWidget:
        parent = QObject.parent(self)
        if APP_DEBUG or TYPE_CHECKING:
            assert isinstance(parent, QWidget)
        return parent

    @property
    def repo(self) -> Repo:
        return self.repoModel.repo

    @property
    def collapseCache(self) -> set[str]:
        return self.collapseCacheLayers[-1]

    def __init__(self, parent=None):
        super().__init__(parent)

        # Initialize collapse cache with an empty permanent layer
        # (i.e. all items start expanded)
        self.collapseCacheLayers = [set()]

        self.clear()

        if APP_DEBUG and HAS_QTEST:
            self.modelTester = QAbstractItemModelTester(self)
            if not APP_TESTMODE:
                warnings.warn("Sidebar model tester enabled. This will SIGNIFICANTLY slow down SidebarModel.rebuild!")

    def clear(self, emitSignals=True):
        # IMPORTANT: Do not clear collapseCache in this function!
        # rebuild() calls clear() but we want collapseCache to persist!

        if emitSignals:
            self.beginResetModel()

        self.repoModel = None
        self.rootNode = SidebarNode(SidebarItem.Root)
        self.nodesByRef = {}
        self._checkedOut = ""
        self._checkedOutUpstream = ""
        self.clearCachedTooltip()

        if emitSignals:
            self.endResetModel()

    def clearCachedTooltip(self):
        self._cachedToolTipIndex = QModelIndex_default
        self._cachedToolTipText = ""

    def cacheToolTip(self, index: QModelIndex, text: str):
        self._cachedToolTipIndex = index
        self._cachedToolTipText = text

    def isHideAllButThisMode(self):
        return bool(self.repoModel.prefs.showPatterns)

    def isExplicitlyShown(self, node: SidebarNode) -> bool:
        return (self.isHideAllButThisMode()
                and node.refMatchingPattern() in self.repoModel.prefs.showPatterns)

    def isExplicitlyHidden(self, node: SidebarNode) -> bool:
        return (not self.isHideAllButThisMode()
                and node.refMatchingPattern() in self.repoModel.prefs.hidePatterns)

    def isImplicitlyHidden(self, node: SidebarNode) -> bool:
        return (node.isLeafBranchKind()
                and node.data in self.repoModel.hiddenRefs
                and node.data not in self.repoModel.prefs.hidePatterns)

    def isAncestryChainExpanded(self, node: SidebarNode):
        # Everything is expanded if collapseCache is empty.
        if not self.collapseCache:
            return True

        # My collapsed state doesn't matter here - it only affects my children.
        # So start looking at my parent.
        node = node.parent
        assert node is not None

        # Walk up parent chain until root index (row -1)
        while node.parent is not None:
            h = node.getCollapseHash()
            if h in self.collapseCache:
                return False
            node = node.parent

        return True

    def cacheNodeCollapsedState(self, node: SidebarNode, collapsed: bool):
        h = node.getCollapseHash()
        if collapsed:
            self.collapseCache.add(h)
        else:
            self.collapseCache.discard(h)

    def refreshRepoName(self):
        if self.rootNode and self.repoModel:
            workdirNode = self.rootNode.findChild(SidebarItem.WorkdirHeader)
            workdirNode.displayName = settings.history.getRepoNickname(self.repo.workdir)

    @benchmark
    def rebuild(self, repoModel: RepoModel):
        self.beginResetModel()

        repo = repoModel.repo

        self.clear(emitSignals=False)
        self.repoModel = repoModel
        self.nodesByRef = {}

        # Pending ref shorthands for _makeRefTreeNodes
        localBranches = []
        remoteBranchesDict: dict[str, list[str]] = {}
        tags = []

        # -----------------------------
        # Set up root nodes
        # -----------------------------
        rootNode = SidebarNode(SidebarItem.Root)
        for eitem in SidebarLayout.RootItems:
            rootNode.appendChild(SidebarNode(eitem))
        uncommittedNode = rootNode.findChild(SidebarItem.UncommittedChanges)
        branchRoot = rootNode.findChild(SidebarItem.LocalBranchesHeader)
        remoteRoot = rootNode.findChild(SidebarItem.RemotesHeader)
        tagRoot = rootNode.findChild(SidebarItem.TagsHeader)
        submoduleRoot = rootNode.findChild(SidebarItem.SubmodulesHeader)
        stashRoot = rootNode.findChild(SidebarItem.StashesHeader)

        self.rootNode = rootNode
        self.nodesByRef[UC_FAKEREF] = uncommittedNode

        self.refreshRepoName()

        # -----------------------------
        # HEAD
        # -----------------------------
        try:
            # Try to get the name of the checked-out branch
            checkedOut = repo.head.name

        except GitError:
            # Unborn HEAD - Get name of unborn branch
            assert repo.head_is_unborn
            target = repo.lookup_reference("HEAD").target
            assert isinstance(target, str), "Unborn HEAD isn't a symbolic reference!"
            target = target.removeprefix(RefPrefix.HEADS)
            node = SidebarNode(SidebarItem.UnbornHead, target)
            branchRoot.appendChild(node)
            self.nodesByRef["HEAD"] = node

        else:
            # It's not unborn
            if checkedOut == 'HEAD':
                # Detached head, leave self._checkedOut blank
                assert repo.head_is_detached
                node = SidebarNode(SidebarItem.DetachedHead, str(repo.head.target))
                branchRoot.appendChild(node)
                self.nodesByRef["HEAD"] = node

            else:
                # We're on a branch
                assert checkedOut.startswith(RefPrefix.HEADS)
                checkedOut = checkedOut.removeprefix(RefPrefix.HEADS)
                self._checkedOut = checkedOut

                # Try to get the upstream
                try:
                    upstream = self.repoModel.upstreams[checkedOut]
                except KeyError:
                    upstream = ""
                self._checkedOutUpstream = upstream

        # -----------------------------
        # Remotes
        # -----------------------------
        for name in repoModel.remotes:
            remoteBranchesDict[name] = []
            node = SidebarNode(SidebarItem.Remote, name)
            remoteRoot.appendChild(node)

        # -----------------------------
        # Refs
        # -----------------------------
        for name in reversed(repoModel.refs):  # reversed because refCache sorts tips by ASCENDING commit time
            prefix, shorthand = RefPrefix.split(name)

            if prefix == RefPrefix.HEADS:
                localBranches.append(shorthand)
                # We're not caching upstreams because it's very expensive to do

            elif prefix == RefPrefix.REMOTES:
                remote, branchName = split_remote_branch_shorthand(shorthand)
                try:
                    remoteBranchesDict[remote].append(branchName)
                except KeyError:
                    warnings.warn(f"SidebarModel: missing remote: {remote}")

            elif prefix == RefPrefix.TAGS:
                tags.append(shorthand)

            elif name == "HEAD" or name.startswith("stash@{"):
                pass  # handled separately

            elif name == UC_FAKEREF:
                pass

            else:
                warnings.warn(f"SidebarModel: unsupported ref prefix: {name}")

        # Populate local branch tree
        self.populateRefNodeTree(localBranches, branchRoot, SidebarItem.LocalBranch, RefPrefix.HEADS, repoModel.prefs.sortBranches)

        # Populate tag tree
        self.populateRefNodeTree(tags, tagRoot, SidebarItem.Tag, RefPrefix.TAGS, repoModel.prefs.sortTags)

        # Populate remote tree
        for remote, branches in remoteBranchesDict.items():
            remoteNode = remoteRoot.findChild(SidebarItem.Remote, remote)
            assert remoteNode is not None
            remotePrefix = f"{RefPrefix.REMOTES}{remote}/"
            self.populateRefNodeTree(branches, remoteNode, SidebarItem.RemoteBranch, remotePrefix, repoModel.prefs.sortRemoteBranches)

        # -----------------------------
        # Stashes
        # -----------------------------
        for i, stashCommitId in enumerate(repoModel.stashes):
            message = repo[stashCommitId].peel(Commit).message
            message = strip_stash_message(message)
            refName = f"stash@{{{i}}}"
            node = SidebarNode(SidebarItem.Stash, str(stashCommitId))
            node.displayName = message
            stashRoot.appendChild(node)
            self.nodesByRef[refName] = node

        # -----------------------------
        # Submodules
        # -----------------------------
        for submoduleKey in repoModel.submodules:
            node = SidebarNode(SidebarItem.Submodule, submoduleKey)
            submoduleRoot.appendChild(node)

            if submoduleKey not in repoModel.initializedSubmodules:
                node.warning = _("Submodule not initialized.")

        # -----------------------------
        # Commit new model
        # -----------------------------
        self.endResetModel()

    def populateRefNodeTree(
            self,
            shorthands: list[str],
            containerNode: SidebarNode,
            kind: SidebarItem,
            refNamePrefix: str,
            sortMode: RefSort = RefSort.UseGlobalPref
    ):
        pendingFolders: dict[str, SidebarNode] = {}

        if sortMode == RefSort.UseGlobalPref:
            sortMode = settings.prefs.refSort
        assert sortMode != RefSort.UseGlobalPref

        shIter: Iterable[str]
        if sortMode == RefSort.TimeAsc:
            shIter = reversed(shorthands)
        elif sortMode == RefSort.AlphaAsc:
            shIter = sorted(shorthands, key=naturalSort)
        elif sortMode == RefSort.AlphaDesc:
            shIter = sorted(shorthands, key=naturalSort, reverse=True)
        else:
            shIter = shorthands

        for sh in shIter:
            if not BRANCH_FOLDERS or "/" not in sh:
                folderNode = containerNode
            else:
                folderName = sh.rsplit("/", 1)[0]
                try:
                    folderNode = pendingFolders[folderName]
                except KeyError:
                    # Create node for folder, but add it to containerNode later
                    # so that all folders are grouped together.
                    folderNode = SidebarNode(SidebarItem.RefFolder, refNamePrefix + folderName)
                    pendingFolders[folderName] = folderNode

            refName = refNamePrefix + sh
            node = SidebarNode(kind, refName)
            folderNode.appendChild(node)
            self.nodesByRef[refName] = node

        for folderName, folderNode in pendingFolders.items():
            parts = folderName.split("/")
            parts.pop()
            while parts:
                parentFolder = "/".join(parts)
                try:
                    parentNode = pendingFolders[parentFolder]
                except KeyError:
                    parts.pop()
                else:
                    parentNode.appendChild(folderNode)
                    folderNode.displayName = folderNode.data.removeprefix(parentNode.data + "/")
                    break
            else:
                folderNode.displayName = folderNode.data.removeprefix(refNamePrefix)
                containerNode.appendChild(folderNode)

    def createIndexFromNode(self, node: SidebarNode) -> QModelIndex:
        index = self.createIndex(node.row, 0, node)
        return index

    def index(self, row: int, column: int, parent: QModelIndex = QModelIndex_default) -> QModelIndex:
        # Return an index given a parent and a row (i.e. child number within parent)

        assert column == 0
        assert row >= 0

        if not parent.isValid():
            parentNode = self.rootNode
        else:
            parentNode = SidebarNode.fromIndex(parent)

        node = parentNode.children[row]
        assert node.row == row

        return self.createIndexFromNode(node)

    # Type checking boilerplate
    @overload
    def parent(self, child: QModelIndex) -> QModelIndex: ...

    # Type checking boilerplate
    @overload
    def parent(self) -> QObject | None: ...

    def parent(self, child: QModelIndex | None = None) -> QModelIndex | QObject | None:
        if child is None:
            return super().parent()

        # Return the parent of the given index
        index = child

        # No repo or root node: no parent
        if not index.isValid():
            return QModelIndex()

        # Get parent node
        node = SidebarNode.fromIndex(index)
        node = node.parent
        assert node is not None

        # Blank index for children of root node
        if node is self.rootNode:
            return QModelIndex()

        return self.createIndexFromNode(node)

    def rowCount(self, parent: QModelIndex = QModelIndex_default) -> int:
        if not parent.isValid():  # root
            node = self.rootNode
        else:
            node = SidebarNode.fromIndex(parent)
        return len(node.children)

    def columnCount(self, parent: QModelIndex = QModelIndex_default) -> int:
        return 1

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if not index.isValid():
            return None

        # Tooltips may show information that is expensive to obtain.
        # Try to reuse any tooltip that we may have cached for the same index.
        if role == Qt.ItemDataRole.ToolTipRole and index == self._cachedToolTipIndex:
            return self._cachedToolTipText

        node = SidebarNode.fromIndex(index)
        assert node is not None

        displayRole = role == Qt.ItemDataRole.DisplayRole
        toolTipRole = role == Qt.ItemDataRole.ToolTipRole
        fontRole = role == Qt.ItemDataRole.FontRole
        refRole = role == SidebarModel.Role.Ref
        iconKeyRole = role == SidebarModel.Role.IconKey

        row = index.row()
        item = node.kind

        if item == SidebarItem.Spacer:
            pass

        elif item == SidebarItem.LocalBranch:
            refName = node.data
            branchName = refName.removeprefix(RefPrefix.HEADS)
            if displayRole:
                if not BRANCH_FOLDERS:
                    return branchName
                return branchName.rsplit("/", 1)[-1]
            elif role == SidebarModel.Role.AheadBehind:
                try:
                    return self.repoModel.aheadBehind[branchName]
                except KeyError:
                    return None
            elif role == SidebarModel.Role.MissingUpstream:
                try:
                    upstream = self.repoModel.upstreams[branchName]
                    if upstream and (RefPrefix.REMOTES + upstream) not in self.repoModel.refs:
                        return upstream
                except KeyError:
                    pass
                return None
            elif refRole:
                return refName
            elif toolTipRole:
                text = "<p style='white-space: pre'>"
                text += _("{0} (local branch)", btag(branchName))
                text += "\n" + self.upstreamToolTip(branchName)
                if branchName == self._checkedOut:
                    checkedOutText = _("(this is the checked-out branch)")
                    text += f"\n{stockIconImgTag('git-head')} HEAD {checkedOutText}"
                text += self.visibilityToolTip(node)
                self.cacheToolTip(index, text)
                return text
            elif iconKeyRole:
                return "git-branch" if branchName != self._checkedOut else "git-head"
            elif fontRole:
                if branchName == self._checkedOut:
                    font = QFont(self._parentWidget.font())
                    font.setBold(True)
                    return font
                else:
                    return None

        elif item == SidebarItem.UnbornHead:
            target = node.data
            if displayRole:
                return _("[unborn]") + " " + target
            elif toolTipRole:
                text = ("<p style='white-space: pre'>"
                        + _("Unborn HEAD: does not point to a commit yet.") + "\n"
                        + _("Local branch {0} will be created when you create the initial commit.", bquo(target)))
                self.cacheToolTip(index, text)
                return text

        elif item == SidebarItem.DetachedHead:
            if displayRole:
                return _("Detached HEAD")
            elif toolTipRole:
                oid = Oid(hex=node.data)
                caption = _("Detached HEAD")
                return f"<p style='white-space: pre'>{caption} @ {shortHash(oid)}"
            elif refRole:
                return "HEAD"
            elif iconKeyRole:
                return "git-head-detached"
            elif fontRole:
                font = QFont(self._parentWidget.font())
                font.setBold(True)
                return font

        elif item == SidebarItem.Remote:
            remoteName = node.data
            if displayRole:
                return remoteName
            elif toolTipRole:
                url = self.repo.remotes[remoteName].url
                skipFetchAll = self.repo.get_remote_skipfetchall(remoteName)
                text = "<p style='white-space: pre'>" + escape(url)
                text += self.visibilityToolTip(node)
                if skipFetchAll:
                    text += "<br>" + _("(Skipped when fetching all remotes.)")
                self.cacheToolTip(index, text)
                return text
            elif iconKeyRole:
                return "git-remote"

        elif item == SidebarItem.RemoteBranch:
            refName = node.data
            shorthand = refName.removeprefix(RefPrefix.REMOTES)
            remoteName, branchName = split_remote_branch_shorthand(shorthand)
            if displayRole:
                if not BRANCH_FOLDERS:
                    return branchName
                return branchName.rsplit("/", 1)[-1]
            elif refRole:
                return refName
            elif toolTipRole:
                text = "<p style='white-space: pre'>"
                text += _("{0} (remote-tracking branch)", btag(shorthand))
                if self._checkedOutUpstream == shorthand:
                    text += "<br><i>" + _("Upstream for the checked-out branch ({0})", hquoe(self._checkedOut)) + "</i>"
                text += self.visibilityToolTip(node)
                self.cacheToolTip(index, text)
                return text
            elif fontRole:
                if self._checkedOutUpstream == shorthand:
                    font = QFont(self._parentWidget.font())
                    font.setBold(True)
                    return font
                else:
                    return None
            elif iconKeyRole:
                return "git-branch"

        elif item == SidebarItem.RefFolder:
            refName = node.data
            if displayRole:
                return node.displayName
            elif toolTipRole:
                prefix, name = RefPrefix.split(refName)
                text = f"<p style='white-space: pre'>{stockIconImgTag('git-folder')} "
                if prefix == RefPrefix.REMOTES:
                    text += _("{0} (remote branch folder)", btag(name))
                elif prefix == RefPrefix.TAGS:
                    text += _("{0} (tag folder)", btag(name))
                else:
                    text += _("{0} (local branch folder)", btag(name))
                text += self.visibilityToolTip(node)
                self.cacheToolTip(index, text)
                return text
            elif iconKeyRole:
                return "git-folder"

        elif item == SidebarItem.Tag:
            refName = node.data
            tagName = refName.removeprefix(RefPrefix.TAGS)
            if displayRole:
                if not BRANCH_FOLDERS:
                    return tagName
                return tagName.rsplit("/", 1)[-1]
            elif refRole:
                return refName
            elif toolTipRole:
                text = "<p style='white-space: pre'>"
                text += _("Tag {0}", bquo(tagName))
                return text
            elif iconKeyRole:
                return "git-tag"

        elif item == SidebarItem.Stash:
            if displayRole:
                return node.displayName
            elif refRole:
                return F"stash@{{{row}}}"
            elif toolTipRole:
                oid = Oid(hex=node.data)
                commit = self.repo.peel_commit(oid)
                dateText = signatureDateFormat(commit.committer, settings.prefs.shortTimeFormat)
                text = "<p style='white-space: pre'>"
                text += f"<b>stash@{{{row}}}</b>: {escape(commit.message)}<br/>"
                text += f"<b>{_('date:')}</b> {escape(dateText)}"
                self.cacheToolTip(index, text)
                return text
            elif iconKeyRole:
                return "git-stash"

        elif item == SidebarItem.Submodule:
            if displayRole:
                return node.data.rsplit("/", 1)[-1]
            elif toolTipRole:
                text = "<p style='white-space: pre'>"
                text += _("{0} (submodule)", f"<b>{escape(node.data)}</b>")
                text += "\n" + _("Workdir: {0}", escape(self.repo.listall_submodules_dict()[node.data]))
                url = self.repo.submodules[node.data].url or _("[not set]")
                text += "\n" + _("URL: {0}", escape(url))
                if node.warning:
                    text += "<br>\u26a0 " + node.warning
                self.cacheToolTip(index, text)
                return text
            elif iconKeyRole:
                return "achtung" if node.warning else "git-submodule"

        elif item == SidebarItem.UncommittedChanges:
            if displayRole:
                changesText = trtables.enum(SidebarItem.UncommittedChanges)
                numUncommittedChanges = self.repoModel.numUncommittedChanges
                if numUncommittedChanges >= 0:
                    ucSuffix = f" ({numUncommittedChanges})"
                    changesText = changesText.replace("\x9C", ucSuffix + "\x9C") + ucSuffix
                return changesText
            elif refRole:
                # Return fake ref so we can select Uncommitted Changes from elsewhere
                return UC_FAKEREF
            elif iconKeyRole:
                return "git-workdir"
            elif toolTipRole:
                text = _("Go to Working Directory")
                text = appendShortcutToToolTipText(text, QKeySequence("Ctrl+G"))
                if self.repoModel.numUncommittedChanges >= 0:
                    text += "\n" + _n("({n} uncommitted change)", "({n} uncommitted changes)",
                                     self.repoModel.numUncommittedChanges)
                self.cacheToolTip(index, text)
                return text

        else:
            if displayRole:
                if item == SidebarItem.WorkdirHeader:
                    return node.displayName
                elif item == SidebarItem.LocalBranchesHeader:
                    return trtables.enum(item)
                else:
                    name = trtables.enum(item)
                    if node.getCollapseHash() in self.collapseCache:
                        name += f" ({len(node.children)})"
                    return name
            elif refRole:
                return ""
            elif fontRole:
                font = self._parentWidget.font()
                font.setWeight(QFont.Weight.DemiBold)
                return font

        return None

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags

        node = SidebarNode.fromIndex(index)

        if node.kind == SidebarItem.Spacer:
            return Qt.ItemFlag.ItemNeverHasChildren

        if node.kind in SidebarLayout.ForceExpand:
            return Qt.ItemFlag.NoItemFlags

        f = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable

        if not node.mayHaveChildren():
            assert not node.children
            f |= Qt.ItemFlag.ItemNeverHasChildren

        return f

    def visibilityToolTip(self, node):
        if self.isExplicitlyShown(node):
            icon = "view-exclusive"
            text = _("Hiding everything but this (middle-click the eye to toggle)")
        elif self.isExplicitlyHidden(node):
            icon = "view-hidden"
            text = _("Hidden (click the eye to toggle)")
        elif self.isImplicitlyHidden(node):
            icon = "view-hidden-indirect"
            text = _("Indirectly hidden")
        else:
            return ""
        return f"<br>{stockIconImgTag(icon)} {text}"

    def upstreamToolTip(self, branchName: str):
        parts = []

        try:
            upstream = self.repoModel.upstreams[branchName]
            parts.append(_("Upstream: {0}", escape(upstream)))
        except KeyError:
            return _("No upstream branch")

        try:
            upstreamTarget = self.repoModel.refs[RefPrefix.REMOTES + upstream]
        except KeyError:
            icon = stockIconImgTag('git-upstream-missing')
            text = _("Upstream missing ({0})", escape(upstream))
            return f"{icon} {text}"

        if upstreamTarget == self.repoModel.refs[RefPrefix.HEADS + branchName]:
            text = _("Up-to-date with upstream")
            parts.append(f" {SYMBOL_UPDOWN} {text}")
        else:
            try:
                ahead, behind = self.repoModel.aheadBehind[branchName]
                if ahead:
                    text = _n("{n} commit ahead", "{n} commits ahead", ahead)
                    parts.append(f" {SYMBOL_AHEAD} {text}")
                if behind:
                    text = _n("{n} commit behind", "{n} commits behind", behind)
                    parts.append(f" {SYMBOL_BEHIND} {text}")
            except KeyError:
                pass

        return "\n".join(parts)




# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import dataclasses
import enum
import logging
import shlex
from collections.abc import Generator
from typing import Any, Literal, ClassVar

from gitfourchette.exttools.toolcommands import ToolCommands
from gitfourchette.forms.askpassdialog import AskpassDialog
from gitfourchette.gitdriver import GitDriver
from gitfourchette.manualgc import gcHint
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.porcelain import ConflictError, MultiFileError, Repo, RepositoryState
from gitfourchette.qt import *
from gitfourchette.repomodel import RepoModel
from gitfourchette.toolbox import *

if TYPE_CHECKING:
    from gitfourchette.repowidget import RepoWidget

logger = logging.getLogger(__name__)


def showConflictErrorMessage(parent: QWidget, exc: ConflictError, opName="Operation"):
    numConflicts = len(exc.conflicts)

    title = _n("Conflicting file", "{n} conflicting files", numConflicts)
    nFilesText = _np("operation conflicts with…", "a file", "{n} files", numConflicts)
    nFilesText = f"<b>{nFilesText}</b>"

    if exc.description == "workdir":
        message = _("Operation {op} conflicts with {files} in the working directory:")
    elif exc.description == "HEAD":
        message = _("Operation {op} conflicts with {files} in the commit at HEAD:")
    else:
        message = _("Operation {op} has caused a conflict with {files} ({exc}):")
    message = message.format(op=bquo(opName), files=nFilesText, exc=exc.description)

    qmb = showWarning(parent, title, message)
    addULToMessageBox(qmb, exc.conflicts)

    if exc.description == "workdir":
        dt = qmb.detailedText()
        dt += _("Before you try again, you should either commit, stash, or discard your changes.")
        qmb.setDetailedText(dt)


def showMultiFileErrorMessage(parent: QWidget, exc: MultiFileError, opName="Operation"):
    details = []

    if exc.message:
        message = exc.message
    else:
        message = _n("Operation {op} ran into an issue with {n} file.",
                     "Operation {op} ran into issues with {n} files.",
                     n=len(exc.file_exceptions), op=hquo(opName))
        if exc.file_successes > 0:
            message += " " + _n("({n} other file was successful.)",
                                "({n} other files were successful.)", n=exc.file_successes)

    for path, error in exc.file_exceptions.items():
        if error:
            details.append(f"<b>{escape(path)}</b>{_(':')} {escape(str(error))}")
        else:
            details.append(escape(path))

    qmb = asyncMessageBox(parent, 'warning', opName, message)
    addULToMessageBox(qmb, details)
    qmb.show()


class TaskPrereqs(enum.IntFlag):
    Nothing = 0
    NoUnborn = enum.auto()
    NoDetached = enum.auto()
    NoConflicts = enum.auto()
    NoCherrypick = enum.auto()
    NoStagedChanges = enum.auto()


class TaskEffects(enum.IntFlag):
    """
    Flags indicating which parts of the UI to refresh
    after a task runs to completion.
    """

    Nothing = 0
    "The task doesn't modify the repository."

    Workdir = enum.auto()
    "The task affects indexed and/or unstaged changes."

    Refs = enum.auto()
    "The task affects branches (local or remote), stashes, or tags."

    Remotes = enum.auto()
    "The task affects remotes registered with this repository."

    Head = enum.auto()
    "The task moves HEAD to a different commit."

    Upstreams = enum.auto()
    "The task affects the upstream of a local branch."

    DefaultRefresh = Workdir | Refs | Remotes | Upstreams
    "Default flags for RefreshRepo"
    # Index is included so the banner can warn about conflicts
    # regardless of what part of the repo is being viewed.


class FlowControlToken:
    """
    Object that can be yielded from `RepoTask.flow()` to control the flow of the coroutine.
    """

    BootstrapFlow: ClassVar[FlowControlToken]

    class Kind(enum.IntEnum):
        ContinueOnUiThread = enum.auto()
        ContinueOnWorkThread = enum.auto()
        WaitUserReady = enum.auto()
        WaitProcessReady = enum.auto()
        InterruptedByException = enum.auto()

    flowControl: Kind
    exception: Exception | None

    def __init__(self, flowControl: Kind = Kind.ContinueOnUiThread, exception=None):
        self.flowControl = flowControl
        self.exception = exception

    def __str__(self):
        return F"FlowControlToken({self.flowControl.name})"


FlowControlToken.BootstrapFlow = FlowControlToken()


class FlowWorkerThread(QThread):
    tokenReady = Signal(FlowControlToken)

    flow: RepoTask.Flow[Any] | None

    @calledFromQThread  # enable code coverage in task threads
    def run(self):
        assert self.flow is not None, "flow not set"
        token = RepoTaskRunner._getNextToken(self.flow)
        self.flow = None
        self.tokenReady.emit(token)


class AbortTask(Exception):
    """ To bail from a coroutine early, we must raise an exception to ensure that
    any active context managers exit deterministically."""
    def __init__(
            self,
            text: str = "",
            icon: MessageBoxIconName = "warning",
            asStatusMessage: bool = False,
            details: str = ""
    ):
        super().__init__(text)
        self.icon = icon
        self.asStatusMessage = asStatusMessage
        self.details = details


class RepoGoneError(FileNotFoundError):
    pass


@dataclasses.dataclass
class TaskEpilog:
    effects: TaskEffects = TaskEffects.Nothing
    """ Which parts of the UI should be refreshed when this task completes. """

    jumpTo: NavLocator = NavLocator.Empty
    """ Jump to this location when this task completes. """

    status: str = ""
    """ Display this message in the status bar when this task completes. """

    def __or__(self, other: TaskEpilog):
        return TaskEpilog(
            self.effects | other.effects,
            self.jumpTo or other.jumpTo,
            self.status or other.status)


class RepoTask(QObject):
    """
    Task that manipulates a repository.
    """

    type Flow[FlowReturnType] = Generator[FlowControlToken, None, FlowReturnType]

    uiReady = Signal()

    repo: Repo

    repoModel: RepoModel

    epilog: TaskEpilog
    """ What to do after this task completes. """

    _currentProcess: QProcess | None
    """ External process that this task is currently waiting on (via flowStartProcess).
    This is only valid in the root task in the stack (non-root tasks are allowed
    to set the root task's current process). """

    _currentFlow: Flow[Any] | None
    _currentIteration: int

    _taskStack: list[RepoTask]
    """ Chain of nested subtasks. The reference to this object is shared by all
    subtasks in the same chain. (This is implemented as a FIFO stack of active
    tasks in the chain of nested 'flowSubtask' calls, including the root task at
    index 0.) """

    _transientSubtaskStatus: str

    @classmethod
    def name(cls) -> str:
        from gitfourchette.tasks.taskbook import TaskBook
        return TaskBook.names.get(cls, cls.__name__)

    @classmethod
    def invoke(cls, invoker: QObject, *args, **kwargs):
        call = TaskInvocation(invoker, cls, *args, **kwargs)
        call.run()

    def __init__(self, parent: QObject):
        super().__init__(parent)
        self.setObjectName(self.__class__.__name__)
        self.repo = None
        self.repoModel = None
        self.epilog = TaskEpilog()
        self._currentFlow = None
        self._currentIteration = 0
        self._currentProcess = None
        self._taskStack = [self]  # will be replaced by shared reference
        self._transientSubtaskStatus = ""
        self._runningOnUiThread = True  # for debugging

    @property
    def rootTask(self) -> RepoTask:
        assert self._taskStack
        return self._taskStack[0]

    @property
    def isRootTask(self) -> bool:
        return self.rootTask is self

    @property
    def currentProcess(self) -> QProcess | None:
        """
        Return the currently-running QProcess in this chain of subtasks, or None
        if no process is running in this chain. There can only be up to a single
        running process per chain, and new subtasks cannot be started while a
        process is running.
        """
        assert not any(t._currentProcess for t in self._taskStack[:-1]), \
            "only the last subtask may own a running process"

        return self._taskStack[-1]._currentProcess

    def parentWidget(self) -> QWidget:
        return findParentWidget(self)

    @property
    def rw(self) -> RepoWidget:  # hack for now - assume parent is a RepoWidget
        parentWidget = self.parentWidget()
        if APP_DEBUG or TYPE_CHECKING:
            from gitfourchette.repowidget import RepoWidget
            assert isinstance(parentWidget, RepoWidget)
        return parentWidget

    def setRepoModel(self, repoModel: RepoModel):
        self.repoModel = repoModel
        self.repo = repoModel.repo

    def __str__(self):
        return self.objectName()

    def isFreelyInterruptible(self) -> bool:
        return False

    def canKill(self, task: RepoTask) -> bool:
        """
        Meant to be overridden by your task.
        Return true if this task is allowed to take precedence over the given running task.
        The default implementation will not attempt to kill any other tasks.
        """
        return False

    def broadcastProcesses(self) -> bool:
        """
        Meant to be overridden by your task.
        Returns whether this task should emit the `processStarted` signal when
        `flowStartProcess` starts a process. This enables automatic progress
        reporting via ProcessDialog.
        """
        return True

    def _isRunningOnAppThread(self):
        return onAppThread() and self._runningOnUiThread

    def flow(self, *args, **kwargs) -> Flow[Any]:
        """
        Generator that performs the task. You can think of this as a coroutine.

        You can control the flow of the coroutine by yielding a `FlowControlToken` object.
        This lets you wait for you user input via dialog boxes, abort the task, or move long
        computations to a separate thread.

        It is recommended to `yield from` one of the `flowXXX` methods instead of instantiating
        a FlowControlToken directly. For example::

            meaning = QInputDialog.getInt(self.parentWidget(), "Hello",
                                          "What's the meaning of life?")
            yield from self.flowEnterWorkerThread()
            expensiveComputationCorrect = meaning == 42
            yield from self.flowEnterUiThread()
            if not expensiveComputationCorrect:
                raise AbortTask("Sorry, computer says no.")

        The coroutine always starts on the UI thread.
        """
        # Dummy yield to make it a generator. You should override this function anyway!
        yield from self.flowEnterUiThread()

    def cleanup(self):
        """
        Clean up any resources used by the task on completion or failure.
        Meant to be overridden by your task.
        Called from UI thread.
        """
        assert onAppThread()

    def onError(self, exc: Exception):
        """
        Report an error to the user if flow() was interrupted by an exception.
        Can be overridden by your task, but you should call super().onError() if you can't handle the exception.
        Called from the UI thread, after cleanup().
        """
        if isinstance(exc, AbortTask):
            pass
        elif isinstance(exc, ConflictError):
            showConflictErrorMessage(self.parentWidget(), exc, self.name())
        elif isinstance(exc, MultiFileError):
            showMultiFileErrorMessage(self.parentWidget(), exc, self.name())
        else:
            message = _("Operation failed: {0}.", escape(self.name()))
            excMessageBox(exc, title=self.name(), message=message, parent=self.parentWidget())

    def prereqs(self) -> TaskPrereqs:
        """
        Can be overridden by your task.
        """
        return TaskPrereqs.Nothing

    def flowEnterWorkerThread(self):
        """
        Move the task to a non-UI thread.
        (Note that the flow always starts on the UI thread.)

        This function is intended to be called by flow() with "yield from".
        """
        assert self._currentFlow is not None
        self._runningOnUiThread = False
        yield FlowControlToken(FlowControlToken.Kind.ContinueOnWorkThread)

    def flowEnterUiThread(self):
        """
        Move the task to the UI thread.
        (Note that the flow always starts on the UI thread.)

        This function is intended to be called by flow() with "yield from".
        """
        assert self._currentFlow is not None
        self._runningOnUiThread = True
        yield FlowControlToken(FlowControlToken.Kind.ContinueOnUiThread)

    def flowSubtask[T: RepoTask](self, subtaskClass: type[T], *args, **kwargs) -> Flow[Any]:
        """
        Run a subtask's flow() method as if it were part of this task.
        Note that if the subtask raises an exception, the root task's flow will be stopped as well.
        You must be on the UI thread before starting a subtask.

        This function is intended to be called by flow() with "yield from".
        """

        assert self._currentFlow is not None
        assert self._isRunningOnAppThread(), "Subtask must be started start on UI thread"

        # To ensure correct deletion of the subtask when we get deleted, we are the subtask's parent
        subtask = subtaskClass(self)
        subtask.setRepoModel(self.repoModel)
        subtask.setObjectName(f"{self.objectName()}:{subtask.objectName()}")

        # Push subtask onto stack
        subtask._taskStack = self._taskStack  # share reference to task stack
        self._taskStack.append(subtask)

        # Get flow generator from subtask
        subtask._currentFlow = subtask.flow(*args, **kwargs)
        assert isinstance(subtask._currentFlow, Generator), "flow() must contain at least one yield statement"

        # Forward coroutine continuation signal
        subtask.uiReady.connect(self.uiReady)

        # Actually perform the subtask
        subtaskResult = yield from subtask._currentFlow

        # Make sure we're back on the UI thread before re-entering the root task
        if not self._isRunningOnAppThread():
            yield FlowControlToken(FlowControlToken.Kind.ContinueOnUiThread)

        # Pop subtask off chain of subtasks
        rc = self._popSubtask()
        assert rc is subtask

        return subtaskResult

    def _popSubtask(self) -> RepoTask:
        assert self._taskStack, "task stack is already empty!"
        assert onAppThread()

        # Pop last subtask off stack
        subtask = self._taskStack.pop()

        # Clear out reference to shared list
        assert subtask._taskStack is self._taskStack, "subtask not in same chain as root task?"
        subtask._taskStack = None

        # Consume transient status
        self._transientSubtaskStatus = (subtask.epilog.status
                                        or subtask._transientSubtaskStatus
                                        or self._transientSubtaskStatus)
        subtask.epilog.status = ""

        # Percolate epilog to caller task
        # Give way to *this* task's epilog (LHS) if anything was set in it
        self.epilog |= subtask.epilog

        # Clean up subtask (on UI thread)
        subtask.cleanup()

        return subtask

    def flowRequestForegroundUi(self) -> Flow[None]:
        """
        Pause the coroutine until the parent widget becomes the foreground tab.
        No-op if the parent widget is already visible or lacks the
        `becameVisible` signal.

        This function is intended to be called by flow() with "yield from".
        """
        assert self._currentFlow is not None
        assert self._isRunningOnAppThread()

        parentWidget = self.parentWidget()
        if parentWidget.isVisible():
            return

        token = FlowControlToken(FlowControlToken.Kind.WaitUserReady)
        try:
            becameVisible: SignalInstance = parentWidget.becameVisible  # type: ignore[attr-defined]
        except AttributeError:
            logger.warning("Widget does not support becameVisible")
            return

        becameVisible.connect(self.uiReady)
        yield token
        becameVisible.disconnect(self.uiReady)

    def flowStartProcess(self, process: QProcess, autoFail=True, stdin: str = "") -> Flow[None]:
        assert self._isRunningOnAppThread(), "start processes from UI thread"
        assert not any(t.currentProcess for t in self._taskStack), \
            "a process is already running in this subtask chain"

        self._currentProcess = process
        processWrapper = ProcessWrapper(process, self)
        processWrapper.continueCoroutine.connect(self.uiReady)

        if self.isFreelyInterruptible():
            # If the process is fast enough, the synchronous waits may cause the
            # coroutine to run as a single uninterruptible block. Don't wait
            # synchronously to give the user a chance to interrupt the task.
            processWrapper.waitForStartedMaxDelay = 0
            processWrapper.waitForFinishedMaxDelay = 0

        try:
            yield from processWrapper.coWaitStart()
            if stdin:
                process.write(stdin.encode("utf-8"))
                # Close the write channel so children that read stdin to EOF
                # (e.g. 'git lfs smudge') don't block forever.
                process.closeWriteChannel()
            yield from processWrapper.coWaitFinished(autoFail)
        finally:
            self._currentProcess = None
            processWrapper.deleteLater()

    def flowCallGit(
            self,
            *args: str,
            customKey="",
            workdir="",
            env: dict[str, str] | None = None,
            autoFail=True,
    ) -> Flow[GitDriver]:
        process = self.createGitProcess(*args, customKey=customKey, workdir=workdir, env=env)
        yield from self.flowStartProcess(process, autoFail=autoFail)
        return process

    def createGitProcess(
            self,
            *args: str,
            customKey="",
            workdir="",
            env: dict[str, str] | None = None,
    ) -> GitDriver:
        from gitfourchette import settings
        from gitfourchette.application import GFApplication
        from gitfourchette.porcelain import GitConfigHelper

        repo = self.repo

        if not workdir:
            workdir = repo.workdir if repo is not None else ""

        # ---------------------------------------------------------------------
        # Prepare environment variables

        env = env or {}
        sshOptions = []

        # Force Git output in English
        env["LC_ALL"] = "C.UTF-8"

        # FLATPAK: Some git commands like verify-commit use git_mkstemp to pass
        # paths to subprocesses like gpg. Files created in the sandbox's /tmp
        # won't be accessible to programs running on the host. (A typical
        # scenario is a sandboxed git that starts gpg on the host.)
        if FLATPAK:
            env["TMPDIR"] = qTempDir()

        # SSH_ASKPASS
        if settings.prefs.ownAskpass:
            env |= AskpassDialog.environmentForChildProcess(settings.prefs.isGitSandboxed())

        # Use internal ssh-agent
        if settings.prefs.ownSshAgent:
            sshAgent = GFApplication.instance().sshAgent
            if sshAgent:
                env |= sshAgent.environment
                sshOptions += ["-o", "AddKeysToAgent=yes"]

        # Custom SSH key file
        if not customKey and self.repoModel:
            customKey = self.repoModel.prefs.customKeyFile
        if customKey:
            sshOptions += ["-i", customKey, "-o", "IdentitiesOnly=yes"]

        # Apply any custom OpenSSH options
        if sshOptions:
            # Get original ssh command
            if repo is not None:
                sshCommand = repo.get_config_value("core.sshCommand")
            else:
                sshCommand = GitConfigHelper.get_default_value("core.sshCommand")
            sshCommand = sshCommand or "/usr/bin/ssh"
            sshCommandTokens = ToolCommands.splitCommandTokens(sshCommand)
            # Add custom options and join back into a string
            sshCommand = shlex.join(sshCommandTokens + sshOptions)
            # Pass to git (note: it's also possible to pass '-c core.sshCommand={sshCommand}')
            env["GIT_SSH_COMMAND"] = sshCommand

        # ---------------------------------------------------------------------
        # Create GitDriver (QProcess)

        process = GitDriver(*args, parent=self)
        if workdir:
            process.setWorkingDirectory(workdir)
        ToolCommands.setQProcessEnvironment(process, env)
        ToolCommands.wrapFlatpakCommand(process)

        return process

    def flowDialog(self, dialog: QDialog, abortTaskIfRejected=True, proceedSignal=None):
        """
        Show a QDialog, then pause the coroutine until it's accepted or rejected.

        If abortTaskIfRejected is True, rejecting the dialog causes the task
        to be aborted altogether.

        This function is intended to be called by flow() with "yield from".
        """

        assert self._currentFlow is not None
        assert self._isRunningOnAppThread()  # we'll touch the UI

        yield from self.flowRequestForegroundUi()

        waitToken = FlowControlToken(FlowControlToken.Kind.WaitUserReady)
        didReject = False

        def onReject():
            nonlocal didReject
            didReject = True

        dialog.rejected.connect(onReject)
        dialog.finished.connect(self.uiReady)
        if proceedSignal:
            proceedSignal.connect(self.uiReady)

        # Only show the dialog if not made visible by the task already
        # (re-showing may reset its position on the screen)
        if not dialog.isVisible():
            dialog.show()

        yield waitToken

        if abortTaskIfRejected and didReject:
            dialog.deleteLater()
            raise AbortTask("")

        dialog.rejected.disconnect(onReject)
        dialog.finished.disconnect(self.uiReady)
        if proceedSignal:
            proceedSignal.disconnect(self.uiReady)

    def flowFileDialog(self, dialog: QFileDialog) -> Flow[str]:
        yield from self.flowDialog(dialog)

        files = dialog.selectedFiles()
        path = files[0]
        dialog.deleteLater()

        return path

    def flowConfirm(
            self,
            title: str = "",
            text: str = "",
            buttonIcon: str = "",
            verb: str = "",
            cancelText: str = "",
            helpText: str = "",
            detailList: list[str] | None = None,
            dontShowAgainKey: str = "",
            canCancel: bool = True,
            icon: MessageBoxIconName | Literal[""] = "",
            checkbox: QCheckBox | None = None,
            actionButton: QPushButton | None = None,
    ):
        """
        Ask the user to confirm the operation via a message box.
        Interrupts flow() if the user denies.

        This function is intended to be called by flow() with "yield from".
        """

        assert self._currentFlow is not None
        assert self._isRunningOnAppThread()  # we'll touch the UI

        if dontShowAgainKey:
            from gitfourchette import settings
            if dontShowAgainKey in settings.prefs.dontShowAgain:
                logger.debug(f"Skipping dontShowAgainMessage: {text}")
                return

        if not title:
            title = self.name()

        if not verb and canCancel:
            verb = title

        buttonMask = QMessageBox.StandardButton.Ok
        if canCancel:
            icon = icon or "question"
            buttonMask |= QMessageBox.StandardButton.Cancel
        else:
            icon = icon or "information"

        qmb = asyncMessageBox(self.parentWidget(), icon, title, text, buttonMask,
                              deleteOnClose=False)

        dontShowAgainCheckBox = None
        if dontShowAgainKey:
            assert not checkbox
            dontShowAgainPrompt = _("Don’t ask me to confirm this again") if canCancel else _("Don’t show this again")
            dontShowAgainCheckBox = QCheckBox(dontShowAgainPrompt, qmb)
            tweakWidgetFont(dontShowAgainCheckBox, 80)
            qmb.setCheckBox(dontShowAgainCheckBox)
        elif checkbox:
            qmb.setCheckBox(checkbox)

        # Using QMessageBox.StandardButton.Ok instead of QMessageBox.StandardButton.Discard
        # so it connects to the "accepted" signal.
        yes: QAbstractButton = qmb.button(QMessageBox.StandardButton.Ok)
        if buttonIcon:
            yes.setIcon(stockIcon(buttonIcon))
        if verb:
            yes.setText(verb)

        if cancelText:
            assert canCancel, "don't set cancelText when canCancel is False!"
            qmb.button(QMessageBox.StandardButton.Cancel).setText(cancelText)

        if actionButton:
            qmb.addButton(actionButton, QMessageBox.ButtonRole.ApplyRole)

        if helpText:
            hintButton = QHintButton(qmb, helpText)
            hintButton.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
            hintButton.setAutoRaise(False)
            qmb.addButton(hintButton, QMessageBox.ButtonRole.HelpRole)
            hintButton.clicked.disconnect()  # Qt internally wires the button to close the dialog; undo that.
            hintButton.connectClicked()

        if detailList:
            addULToMessageBox(qmb, detailList)

        yield from self.flowDialog(qmb)
        result = qmb.result()

        if dontShowAgainKey and dontShowAgainCheckBox.isChecked():
            from gitfourchette import settings
            settings.prefs.dontShowAgain.append(dontShowAgainKey)
            settings.prefs.setDirty()

        qmb.deleteLater()
        return result

    def checkPrereqs(self, prereqs=TaskPrereqs.Nothing):
        if prereqs == TaskPrereqs.Nothing:
            prereqs = self.prereqs()
        repo = self.repo

        if TaskPrereqs.NoConflicts in prereqs and repo.any_conflicts:
            raise AbortTask(_("Fix merge conflicts before performing this action."))

        if TaskPrereqs.NoUnborn in prereqs and repo.head_is_unborn:
            raise AbortTask(paragraphs(
                _("There are no commits in this repository yet."),
                _("Create the initial commit in this repository before performing this action.")))

        if TaskPrereqs.NoDetached in prereqs and repo.head_is_detached:
            raise AbortTask(paragraphs(
                _("You are in “detached HEAD” state."),
                _("Please switch to a local branch before performing this action.")))

        if TaskPrereqs.NoCherrypick in prereqs and repo.state() == RepositoryState.CHERRYPICK:
            raise AbortTask(paragraphs(
                _("You are in the middle of a cherry-pick."),
                _("Before performing this action, conclude the cherry-pick.")))

        if TaskPrereqs.NoStagedChanges in prereqs and repo.any_staged_changes:
            raise AbortTask(paragraphs(
                _("You have staged changes."),
                _("Before performing this action, commit your changes or stash them.")))


class RepoTaskRunner(QObject):
    ForceSerial = APP_NOTHREADS
    """
    Force tasks to run synchronously on the UI thread.
    Useful for debugging.
    Can be forced with command-line switch "--no-threads".
    """

    ready = Signal()
    """
    Emitted after finishing a task (or a series of tasks) and no other tasks
    are queued.
    """

    progress = Signal(str, bool)
    repoGone = Signal()
    requestAttention = Signal()
    processStarted = Signal(QProcess, str)

    repoModel: RepoModel | None

    pendingEpilog: TaskEpilog

    _workerThread: FlowWorkerThread
    "Thread that can execute non-UI sections of the current task's coroutine."

    _interruptCurrentTask: bool
    "Flag to interrupt the current task."

    _currentTask: RepoTask | None
    "Task that is currently running"

    _pendingTask: RepoTask | None
    "Task that is queued to run after _currentTask"

    _currentTaskBenchmark: Benchmark
    "Context manager"

    _queueTokens: bool
    _tokenQueue: list[FlowControlToken]

    def __init__(self, parent: QObject):
        super().__init__(parent)
        self.setObjectName("RepoTaskRunner")
        self._currentTask = None
        self._pendingTask = None
        self._currentTaskBenchmark = Benchmark("???")
        self._interruptCurrentTask = False
        self.repoModel = None
        self.pendingEpilog = TaskEpilog()

        self._workerThread = FlowWorkerThread(self)
        self._workerThread.flow = None
        self._workerThread.tokenReady.connect(self._continueFlow)

        self._queueTokens = False
        self._tokenQueue = []

    @property
    def currentTask(self):
        return self._currentTask

    def isBusy(self) -> bool:
        return (self._currentTask is not None
                or self._pendingTask is not None
                or self._workerThread.isRunning())

    def prepareForDeletion(self):
        self.killCurrentTask()
        self.joinKilledTask()
        self.repoModel = None

    def killCurrentTask(self):
        """
        Interrupt current task next time it yields a FlowControlToken.
        Also terminate any running process associated with the current task.

        The task will not die immediately. Use joinKilledTask() after killing
        the task to block the current thread until the task runner is empty.

        Note that this will clear the pending task as well.

        No-op if no task is running.
        """

        # When we want to kill a task, invalidate the pending task as well.
        self._releasePendingTask()

        if self._currentTask:
            self._interruptCurrentTask = True
            ToolCommands.terminatePlus(self._currentTask.currentProcess)

    def killTaskClass(self, taskClass: type[RepoTask]):
        """
        Interrupt any current or pending task of the given class.
        """
        # Purge pending task first
        if isinstance(self._pendingTask, taskClass):
            self._releasePendingTask()
        # Then kill current task
        if isinstance(self._currentTask, taskClass):
            self.killCurrentTask()

    def joinKilledTask(self):
        """
        Block UI thread until the current task that is being interrupted is dead.
        Returns immediately if no task is being interrupted.
        """
        assert onAppThread()
        while self._interruptCurrentTask:
            QThread.yieldCurrentThread()
            QThread.msleep(30)
            flags = QEventLoop.ProcessEventsFlag.ExcludeUserInputEvents
            flags |= QEventLoop.ProcessEventsFlag.WaitForMoreEvents
            QApplication.processEvents(flags, 30)
        self.joinWorkerThread()
        assert not self.isBusy(), f"still busy - {self._currentTask} {self._pendingTask} {self._workerThread.isRunning()}"

    def joinWorkerThread(self):
        assert onAppThread()
        if self._workerThread.isRunning():
            self._workerThread.wait()
        assert not self._workerThread.isRunning()
        assert not self._workerThread.flow

    def put(self, call: TaskInvocation):
        assert onAppThread()

        task = call.taskClass(self)
        if self.repoModel:
            task.setRepoModel(self.repoModel)

        # Get flow generator
        task._currentFlow = task.flow(*call.taskArgs, **call.taskKwargs)
        assert isinstance(task._currentFlow, Generator), "flow() must contain at least one yield statement"

        # If there's any pending task, replace it.
        self._releasePendingTask()
        assert not self._pendingTask

        if not self._currentTask:
            # We'll be able to start immediately.
            pass
        elif self._currentTask.isFreelyInterruptible() or task.canKill(self._currentTask):
            # Interrupt the current task.
            logger.debug(f"Task {task} killed task {self._currentTask}")
            self.killCurrentTask()
        else:
            # Enqueue after the current task.
            logger.debug(f"Task {task} enqueued after {self._currentTask}")

        # Queue up this task.
        self._pendingTask = task

        # If the current task slot is vacant at this point, start immediately.
        if not self._currentTask:
            self._startPendingTask()

    def _startPendingTask(self):
        task = self._pendingTask
        assert not self._currentTask, f"cannot start pending task {task} while {self._currentTask} is still current"

        if not task:
            return

        assert task._currentIteration == 0, "pending task isn't supposed to have started yet"
        assert task._currentFlow
        assert task.isRootTask
        assert not self._interruptCurrentTask, "interrupt flag not reset"

        self._currentTask = task
        self._pendingTask = None

        logger.debug(f">>> {task}")
        self._currentTaskBenchmark.name = str(task)
        self._currentTaskBenchmark.__enter__()

        # When the task is ready, continue the coroutine
        task.uiReady.connect(self._continueFlow)

        # Check task prerequisites
        try:
            task.checkPrereqs()
        except AbortTask as abort:
            self.reportAbortTask(task, abort)
            self._releaseCurrentTask()
            return

        # Start the coroutine
        self._continueFlow()

    def _continueFlow(self, token: FlowControlToken = FlowControlToken.BootstrapFlow):
        if self._queueTokens:
            self._tokenQueue.append(token)
            return

        while token is not None:
            token = self._processToken(token)

    def _processToken(self, token: FlowControlToken) -> FlowControlToken | None:
        assert not isinstance(token, Generator), \
            "You're trying to yield a nested generator. Did you mean 'yield from'?"
        assert isinstance(token, FlowControlToken), \
            f"In a RepoTask coroutine, you can only yield FlowControlToken. You yielded: {type(token).__name__}"
        assert onAppThread(), "_processToken must be called on UI thread"

        task = self._currentTask
        assert task is not None
        task._currentIteration += 1

        # Let worker thread wrap up
        self.joinWorkerThread()

        # Wrap up if we've been interrupted
        if self._interruptCurrentTask:
            self._interruptCurrentTask = False
            self._onTaskInterrupted(StopIteration("task was killed"))
            return None

        # ---------------------------------------------------------------------
        # Process the token

        flow = task._currentFlow
        assert flow is not None

        tk = token.flowControl
        TK = FlowControlToken.Kind

        if tk == TK.ContinueOnUiThread:
            # Get next continuation token on this thread then loop to beginning of _continueFlow.
            token = RepoTaskRunner._getNextToken(flow)
            assert token is not None, "Do not yield None from a RepoTask coroutine"
            return token

        elif tk == TK.WaitUserReady:
            # When user is ready, task.uiReady will fire, and we'll re-enter _continueFlow.
            self.progress.emit("", False)
            self.requestAttention.emit()

        elif tk == TK.WaitProcessReady:
            busyMessage = _("Busy: {0}…", task.name())
            self.progress.emit(busyMessage, True)

            # Broadcast process start at most once
            if task.broadcastProcesses():
                process = task.currentProcess
                try:
                    _dummy = process._repoTaskBroadcastYet  # type: ignore[attr-defined]
                except AttributeError:
                    process._repoTaskBroadcastYet = True  # type: ignore[attr-defined]
                    self.processStarted.emit(process, task.name())

            if RepoTaskRunner.ForceSerial:  # In unit tests, block until process has completed
                return self._waitForNextToken()

        elif tk == TK.ContinueOnWorkThread:
            busyMessage = _("Busy: {0}…", task.name())
            self.progress.emit(busyMessage, True)

            # FlowWorkerThread.run() is a wrapper around `next(flow)`.
            # It will eventually re-enter _continueFlow.
            assert not self._workerThread.isRunning()
            self._workerThread.flow = flow
            self._workerThread.start()

            if RepoTaskRunner.ForceSerial:  # In unit tests, block until threaded workload has completed
                return self._waitForNextToken()

        elif tk == TK.InterruptedByException:
            exception = token.exception
            assert exception is not None, "FlowControlToken(InterruptedByException) must provide an exception!"

            self._onTaskInterrupted(exception)

        else:
            raise NotImplementedError(f"Unsupported FlowControlToken {token.flowControl}")

        return None

    def _waitForNextToken(self) -> FlowControlToken:
        """
        Block caller until the next FlowControlToken is ready to be processed.
        Useful to wait for a process to complete, etc.

        Used only in ForceSerial mode to ease the writing of unit tests in a
        "sequential" style.
        """

        assert onAppThread()
        assert not self._queueTokens
        assert not self._tokenQueue

        self._queueTokens = True
        deadline = QDeadlineTimer(5_000)

        while not self._tokenQueue:
            QApplication.processEvents()
            if deadline.hasExpired():  # pragma: no cover
                raise TimeoutError("timed out while waiting for next token")

        self._queueTokens = False
        token = self._tokenQueue.pop(0)
        assert not self._tokenQueue
        return token

    def _onTaskInterrupted(self, exception: Exception):
        task = self._currentTask
        assert task
        assert not self._workerThread.isRunning()

        # Stop tracking this task
        self._releaseCurrentTask()
        assert self._currentTask is None

        ranToCompletion = False

        if isinstance(exception, StopIteration):
            # No more steps in the flow. Task completed successfully.
            ranToCompletion = True
        elif isinstance(exception, AbortTask):
            # Controlled exit, show message (if any)
            self.reportAbortTask(task, exception)
            task.onError(exception)  # Also let task clean up after itself
        elif isinstance(exception, RepoGoneError):
            # Repo directory vanished
            self.repoGone.emit()
        else:
            # Run task's error callback
            task.onError(exception)

        # Accumulate epilog (give way to task epilog over current epilog)
        task.epilog.status = task.epilog.status or task._transientSubtaskStatus
        self.pendingEpilog = task.epilog | self.pendingEpilog

        if not ranToCompletion:
            # Task aborted before completion, discard pending task
            self._releasePendingTask()
        else:
            # Queue up pending task, if any, before postTask
            self._startPendingTask()

        task.deleteLater()

        if self.isBusy():  # There's a pending task
            return

        # Tell owner we're ready for another task
        self.ready.emit()

        if not self.isBusy():  # Owner hasn't chained some other task
            # End of task chain
            # Manual GC: Now's an opportune time to collect garbage
            gcHint()

            # Consume & report pending status
            status = self.pendingEpilog.status
            if status:
                self.pendingEpilog.status = ""
                self.progress.emit(status, False)

    def consumePendingEffectsAndLocator(self) -> tuple[TaskEffects, NavLocator]:
        effects, jumpTo = self.pendingEpilog.effects, self.pendingEpilog.jumpTo
        self.pendingEpilog.effects = TaskEffects.Nothing
        self.pendingEpilog.jumpTo = NavLocator.Empty
        return effects, jumpTo

    @staticmethod
    def _getNextToken(flow: RepoTask.Flow[Any]) -> FlowControlToken:
        try:
            token = next(flow)
        except BaseException as exception:
            token = FlowControlToken(FlowControlToken.Kind.InterruptedByException, exception)
        return token

    def _releaseCurrentTask(self):
        task = self._currentTask

        logger.debug(f"<<< {task} ({self._currentTaskBenchmark.elapsed()*1000:.1f} ms)")
        self.progress.emit("", False)
        self._currentTaskBenchmark.__exit__(None, None, None)

        assert onAppThread()
        assert task is self._currentTask
        assert task.isRootTask

        # Clean up all tasks in the stack (remember, we're the root stack)
        assert task in task._taskStack
        while task._taskStack:
            task._popSubtask()

        assert task._taskStack is None, "popping all subtasks should have cleared reference to chain"

        task.uiReady.disconnect()

        task._currentFlow = None
        self._currentTask = None

    def _releasePendingTask(self):
        if not self._pendingTask:
            return

        logger.debug(f"Pending task {self._pendingTask} discarded")
        assert self._pendingTask._currentIteration == 0, "pending task isn't supposed to have started yet"
        self._pendingTask.deleteLater()
        self._pendingTask = None

    def reportAbortTask(self, task: RepoTask, exception: AbortTask):
        message = str(exception)
        if message and exception.asStatusMessage:
            self.progress.emit("\u26a0 " + message, False)
        elif message:
            parentWidget = self.parent()
            assert isinstance(parentWidget, (QWidget, None))
            qmb = asyncMessageBox(parentWidget, exception.icon, task.name(), message)
            if exception.details:
                qmb.setDetailedText(exception.details)
            qmb.show()


class ProcessWrapper(QObject):
    continueCoroutine = Signal()

    SynchronousMaxDelay = 30
    """
    Skip the coroutine pause/continuation overhead if the process is able to
    spawn within this delay (in milliseconds). This may reduce the time to
    complete the coroutine, but the UI will freeze during this delay.
    """

    def __init__(self, process: QProcess, parent):
        super().__init__(parent)
        self.process = process
        self._didStart = False

        self.waitForStartedMaxDelay = ProcessWrapper.SynchronousMaxDelay
        self.waitForFinishedMaxDelay = ProcessWrapper.SynchronousMaxDelay
        if WINDOWS:  # On Windows, git needs more time to finish
            self.waitForFinishedMaxDelay = 0

    def _onStarted(self):
        self._didStart = True

    def coWaitStart(self):
        logger.info(self.formatCommand(simple=True))

        process = self.process
        process.start()

        # Attempt synchronous start
        if process.waitForStarted(self.waitForStartedMaxDelay):
            self._didStart = True
            return

        with QSignalConnectContext(process.started, self._onStarted):
            if not self._didStart and process.state() == QProcess.ProcessState.Starting:
                # Pause coroutine until process emits either started or errorOccurred
                with (
                    QSignalConnectContext(process.started, self.continueCoroutine),
                    QSignalConnectContext(process.errorOccurred, self.continueCoroutine),
                ):
                    yield FlowControlToken(FlowControlToken.Kind.WaitProcessReady)

        if not self._didStart:
            if isinstance(process, GitDriver):
                message = _("Couldn’t start Git ({0}).", process.error())
            else:
                message = _("Couldn’t start process ({0}).", process.error())
            message += f"<p style='font-size: small'><code>{escape(self.formatCommand(simple=True))}</code><br>"
            raise AbortTask(message)

    def coWaitFinished(self, autoFail=True):
        process = self.process
        assert self._didStart

        if QT5 and process.state() == QProcess.ProcessState.NotRunning:  # pragma: no cover
            # Qt 5 may enter NotRunning immediately in ForceSerial mode
            pass
        else:
            assert process.state() == QProcess.ProcessState.Running

            # Attempt synchronous wait
            if not process.waitForFinished(self.waitForFinishedMaxDelay):
                # The process is taking a while to finish.
                # Pause the coroutine until the process exits the Running state.
                with QSignalConnectContext(process.stateChanged, self.continueCoroutine):
                    yield FlowControlToken(FlowControlToken.Kind.WaitProcessReady)

        assert process.state() != QProcess.ProcessState.Running
        exitCode = process.exitCode()

        if autoFail and exitCode != 0:
            if isinstance(process, GitDriver):
                message = process.htmlErrorText()
            else:
                simpleCommandLine = self.formatCommand(simple=True)
                message = _("Process {0} exited with code {1}.", escape(process.program()), exitCode)
                message += f"<p style='font-size: small'>{escape(simpleCommandLine)}</p>"

                stderr = process.readAllStandardError().data().decode(errors="replace")
                if stderr.strip():
                    message += f"<p style='white-space: pre-wrap'>{escape(stderr)}</p>"

            raise AbortTask(message, details=self.formatCommand())

    def formatCommand(self, simple=False):
        process = self.process
        base = [process.program()] + process.arguments()
        if simple or FLATPAK:
            return shlex.join(base)
        else:
            envStrs = [f"{k}={v}" for k, v in ToolCommands.filterQProcessEnvironment(process).items()]
            return shlex.join(envStrs + base)


class TaskInvocation:
    invoker: QObject
    taskClass: type[RepoTask]
    taskArgs: tuple
    taskKwargs: dict[str, Any]

    def __init__(self, invoker: QObject, taskClass: type[RepoTask], *args, **kwargs):
        assert issubclass(taskClass, RepoTask)
        self.invoker = invoker
        self.taskClass = taskClass
        self.taskArgs = args
        self.taskKwargs = kwargs

    def __repr__(self):
        return f"{self.__class__.__name__}({self.taskClass.__name__})"

    def run(self) -> RepoTaskRunner | None:
        assert isinstance(self.invoker, QObject)
        if self.invoker.signalsBlocked():
            logger.debug(f"Ignoring {self!r} from invoker with blocked signals: " +
                         (self.invoker.objectName() or self.invoker.__class__.__name__))
            return None

        # Find TaskRunner in QObject hierarchy
        runner: RepoTaskRunner
        obj = self.invoker
        if isinstance(obj, RepoTaskRunner):
            runner = obj
        else:
            while obj is not None:
                try:
                    runner = obj.taskRunner  # type: ignore[attr-defined]
                    assert isinstance(runner, RepoTaskRunner)
                    break
                except AttributeError:
                    obj = obj.parent()
            else:
                raise AssertionError("Could not find RepoTaskRunner")

        runner.put(self)
        return runner



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.newtagdialog import populateRemoteComboBox
from gitfourchette.forms.ui_deletetagdialog import Ui_DeleteTagDialog
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class DeleteTagDialog(QDialog):
    def __init__(
            self,
            tagName: str,
            target: str,
            targetSubtitle: str,
            remotes: list[str],
            parent=None):

        super().__init__(parent)

        self.ui = Ui_DeleteTagDialog()
        self.ui.setupUi(self)

        formatWidgetText(self.ui.label, bquo(tagName))

        okButton = self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        okButton.setIcon(stockIcon("SP_DialogDiscardButton"))
        okCaptions = [_("&Delete Locally"), _("&Delete Locally && Remotely")]
        self.ui.pushCheckBox.toggled.connect(lambda push: okButton.setText(okCaptions[push]))

        populateRemoteComboBox(self.ui.remoteComboBox, remotes)

        # Prime enabled state
        self.ui.pushCheckBox.click()
        self.ui.pushCheckBox.click()
        if not remotes:
            self.ui.pushCheckBox.setChecked(False)
            self.ui.pushCheckBox.setEnabled(False)

        convertToBrandedDialog(
            self,
            _("Delete tag {0}", tquo(tagName)),
            _("Tagged commit: {0}", target) + " – " + tquo(targetSubtitle))

        packDialog(self, lockHeight=True)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations  # TODO: Remove once we can drop support for Python <= 3.13

import enum

from gitfourchette.qt import *
from gitfourchette.sidebar.sidebarmodel import SidebarNode, SidebarModel, SidebarItem, SidebarLayout, SYMBOL_AHEAD, SYMBOL_BEHIND
from gitfourchette.toolbox import stockIcon, FittedText
from gitfourchette.toolbox.recolorsvgiconengine import RecolorSvgIconEngine

if TYPE_CHECKING:
    from gitfourchette.sidebar.sidebar import Sidebar

PE_EXPANDED = QStyle.PrimitiveElement.PE_IndicatorArrowDown
PE_COLLAPSED = QStyle.PrimitiveElement.PE_IndicatorArrowRight

# These metrics are a good compromise for Breeze, macOS, and Fusion.
EXPAND_TRIANGLE_WIDTH = 6
PADDING = 4
EYE_WIDTH = 16


class SidebarClickZone(enum.IntEnum):
    Invalid = 0
    Select = 1
    Expand = 2
    Hide = 3


class SidebarDelegate(QStyledItemDelegate):
    """
    Draws custom tree expand/collapse indicator arrows,
    and hide/show icons.
    """

    sidebar: Sidebar

    def __init__(self, parent: Sidebar):
        super().__init__(parent)
        self.sidebar = parent

    @staticmethod
    def unindentRect(item: SidebarItem, rect: QRect, indentation: int):
        if item not in SidebarLayout.UnindentItems:
            return
        unindentLevels = SidebarLayout.UnindentItems[item]
        unindentPixels = unindentLevels * indentation
        return rect.adjust(unindentPixels, 0, 0, 0)

    @staticmethod
    def getClickZone(node: SidebarNode, rect: QRect, x: int):
        if node.kind == SidebarItem.Spacer:
            return SidebarClickZone.Invalid
        elif node.mayHaveChildren() and x < rect.left():
            return SidebarClickZone.Expand
        elif node.canBeHidden() and x > rect.right() - EYE_WIDTH - PADDING:
            return SidebarClickZone.Hide
        else:
            return SidebarClickZone.Select

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex):
        node = self.sidebar.filterIndexToNode(index)
        assert node.parent is not None, "can't paint root node"

        sidebarModel = self.sidebar.sidebarModel
        view = self.sidebar
        assert view is option.widget

        style: QStyle = view.style()
        isActive = bool(option.state & QStyle.StateFlag.State_Active)
        isSelected = bool(option.state & QStyle.StateFlag.State_Selected)
        mouseOver = bool(option.state & QStyle.StateFlag.State_Enabled) and bool(option.state & QStyle.StateFlag.State_MouseOver)
        colorGroup = QPalette.ColorGroup.Active if isActive else QPalette.ColorGroup.Inactive

        isExplicitlyShown = False
        isExplicitlyHidden = False
        isImplicitlyHidden = False
        isHideAllButThisMode = sidebarModel.isHideAllButThisMode()
        makeRoomForEye = False
        if node.canBeHidden():
            isExplicitlyShown = sidebarModel.isExplicitlyShown(node)
            isExplicitlyHidden = sidebarModel.isExplicitlyHidden(node)
            isImplicitlyHidden = isExplicitlyHidden or sidebarModel.isImplicitlyHidden(node)
            makeRoomForEye = mouseOver or isExplicitlyShown or isExplicitlyHidden or isImplicitlyHidden

        painter.save()

        if node.kind == SidebarItem.Spacer:
            mouseOver = False
            option.state &= ~QStyle.StateFlag.State_MouseOver

            r = QRect(option.rect)
            r.setLeft(0)
            middle = r.top() + int(r.height() / 2)

            tc1 = option.palette.color(colorGroup, QPalette.ColorRole.WindowText)
            tc2 = QColor(tc1)
            tc1.setAlpha(0)
            tc2.setAlpha(33)
            lineGradient = QLinearGradient(r.left() + PADDING, middle, r.right() - PADDING, middle)
            lineGradient.setColorAt(0, tc1)
            lineGradient.setColorAt(.2, tc2)
            lineGradient.setColorAt(1-.2, tc2)
            lineGradient.setColorAt(1, tc1)

            painter.save()
            painter.setPen(QPen(lineGradient, 1))
            painter.drawLine(r.left() + PADDING, middle, r.right() - PADDING, middle)
            painter.restore()

        # Unindent rect
        SidebarDelegate.unindentRect(node.kind, option.rect, view.indentation())

        # Draw expand/collapse triangle.
        if node.mayHaveChildren() and not node.wantForceExpand():
            opt2 = QStyleOptionViewItem(option)
            opt2.rect.adjust(-(EXPAND_TRIANGLE_WIDTH + PADDING), 0, 0, 0)  # args must be integers for pyqt5!
            opt2.rect.setWidth(EXPAND_TRIANGLE_WIDTH)

            # See QTreeView::drawBranches() in qtreeview.cpp for other interesting states
            opt2.state &= ~QStyle.StateFlag.State_MouseOver
            arrowPrimitive = PE_EXPANDED if view.isExpanded(index) else PE_COLLAPSED
            style.drawPrimitive(arrowPrimitive, opt2, painter, view)

        # Draw control background
        style.drawControl(QStyle.ControlElement.CE_ItemViewItem, option, painter, option.widget)

        # Adjust contents
        option.rect.adjust(PADDING, 0, -PADDING, 0)

        # Set highlighted text color if this item is selected
        iconMode = QIcon.Mode.Normal
        if isSelected:
            penColor = option.palette.color(colorGroup, QPalette.ColorRole.HighlightedText)
            iconMode = QIcon.Mode.Selected if isActive else QIcon.Mode.SelectedInactive  # type: ignore[attr-defined]
        elif not node.parent.parent and node.kind != SidebarItem.UncommittedChanges:
            penColor = option.palette.color(colorGroup, QPalette.ColorRole.WindowText)
            penColor.setAlphaF(.66)
        else:
            penColor = option.palette.color(colorGroup, QPalette.ColorRole.WindowText)
        painter.setPen(penColor)

        # Draw decoration icon
        iconWidth = option.decorationSize.width()
        iconKey = index.data(SidebarModel.Role.IconKey)
        if iconKey:
            r = QRect(option.rect)
            r.setWidth(iconWidth)
            icon = stockIcon(iconKey)
            icon.paint(painter, r, option.decorationAlignment, mode=iconMode)
            option.rect.adjust(r.width() + PADDING*150//100, 0, 0, 0)

        # Prepare text
        textRect = QRect(option.rect)
        if makeRoomForEye:
            textRect.adjust(0, 0, -EYE_WIDTH, 0)

        font: QFont = index.data(Qt.ItemDataRole.FontRole) or option.font
        baseFontSize = font.pointSizeF()

        # Draw ahead/behind/missing upstream indicators
        missingUpstream = index.data(SidebarModel.Role.MissingUpstream)
        aheadBehind = index.data(SidebarModel.Role.AheadBehind)
        if makeRoomForEye:
            # No upstream indicators
            pass

        elif missingUpstream:
            r = QRect(option.rect)
            r.setLeft(textRect.right() - EYE_WIDTH)
            unpluggedIcon = stockIcon("git-upstream-missing")
            unpluggedIcon.paint(painter, r, mode=iconMode)
            # Clip rect
            textRect.setRight(r.left())

        elif aheadBehind:
            a, b = aheadBehind

            # Set a smaller font
            font.setPointSizeF(baseFontSize * (.67 if a and b else .75))
            painter.setFont(font)
            metrics = painter.fontMetrics()

            textA = f" {a} {SYMBOL_AHEAD}" if a else ""
            textB = f" {b} {SYMBOL_BEHIND}" if b else ""
            advanceA = metrics.horizontalAdvance(textA) if a else 0
            advanceB = metrics.horizontalAdvance(textB) if b else 0

            AF = Qt.AlignmentFlag
            if not isSelected:
                painter.setPen(RecolorSvgIconEngine.IconColors.mainColor)
            painter.drawText(textRect, AF.AlignRight | (AF.AlignTop if b else AF.AlignVCenter) , textA)
            painter.drawText(textRect, AF.AlignRight | (AF.AlignBottom if a else AF.AlignVCenter), textB)

            # Restore font size
            font.setPointSizeF(baseFontSize)
            painter.setPen(penColor)

            # Clip rect
            textRect.setRight(textRect.right() - max(advanceA, advanceB))

        # Draw text
        painter.setFont(font)
        fullText = index.data(Qt.ItemDataRole.DisplayRole)
        isCannedString = node.kind <= SidebarItem.SubmodulesHeader
        if not isCannedString:
            FittedText.draw(painter, textRect, option.displayAlignment, fullText, option.textElideMode)
        else:
            text = painter.fontMetrics().elidedText(fullText, Qt.TextElideMode.ElideMiddle, textRect.width())
            painter.drawText(textRect, option.displayAlignment, text)

        # Draw eye
        if makeRoomForEye:
            r = QRect(option.rect)
            r.setLeft(textRect.right())
            r.setWidth(EYE_WIDTH)
            if isExplicitlyShown or mouseOver and isHideAllButThisMode:
                eyeIconName = "view-exclusive"
            elif isExplicitlyHidden or (isImplicitlyHidden and isHideAllButThisMode):
                eyeIconName = "view-hidden"
            elif isImplicitlyHidden:
                eyeIconName = "view-hidden-indirect"
            else:
                eyeIconName = "view-visible"
            unpluggedIcon = stockIcon(eyeIconName)
            unpluggedIcon.paint(painter, r, mode=iconMode)

        painter.restore()
